################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.fn_execution.embedded.operations import (OneInputFunctionOperation,
                                                      TwoInputFunctionOperation)

from pyflink.fn_execution.embedded.converters import from_type_info_proto, from_schema_proto


def pare_user_defined_data_stream_function_proto(proto):
    from pyflink.fn_execution import flink_fn_execution_pb2
    serialized_fn = flink_fn_execution_pb2.UserDefinedDataStreamFunction()
    serialized_fn.ParseFromString(proto)
    return serialized_fn


def parse_coder_proto(proto):
    from pyflink.fn_execution import flink_fn_execution_pb2

    coder = flink_fn_execution_pb2.CoderInfoDescriptor()
    coder.ParseFromString(proto)
    return coder


def parse_function_proto(proto):
    from pyflink.fn_execution import flink_fn_execution_pb2
    serialized_fn = flink_fn_execution_pb2.UserDefinedFunctions()
    serialized_fn.ParseFromString(proto)
    return serialized_fn


def create_scalar_operation_from_proto(proto,
                                       input_coder_info,
                                       output_coder_into,
                                       one_arg_optimization=False,
                                       one_result_optimization=False):
    from pyflink.fn_execution.table.operations import ScalarFunctionOperation

    serialized_fn = parse_function_proto(proto)

    input_data_converter = (
        from_schema_proto(
            parse_coder_proto(input_coder_info).flatten_row_type.schema,
            one_arg_optimization))

    output_data_converter = (
        from_schema_proto(
            parse_coder_proto(output_coder_into).flatten_row_type.schema,
            one_result_optimization))

    scalar_operation = ScalarFunctionOperation(
        serialized_fn, one_arg_optimization, one_result_optimization)

    process_element_func = scalar_operation.process_element

    def process_element(value):
        actual_value = input_data_converter.to_internal(value)
        result = process_element_func(actual_value)
        return output_data_converter.to_external(result)

    scalar_operation.process_element = process_element
    return scalar_operation


def create_table_operation_from_proto(proto, input_coder_info, output_coder_into):
    from pyflink.fn_execution.table.operations import TableFunctionOperation

    serialized_fn = parse_function_proto(proto)

    input_data_converter = (
        from_schema_proto(parse_coder_proto(input_coder_info).flatten_row_type.schema))
    output_data_converter = (
        from_schema_proto(parse_coder_proto(output_coder_into).flatten_row_type.schema))

    table_operation = TableFunctionOperation(serialized_fn)

    process_element_func = table_operation.process_element

    def process_element(value):
        actual_value = input_data_converter.to_internal(value)
        results = process_element_func(actual_value)
        for result in results:
            yield output_data_converter.to_external(result)

    table_operation.process_element = process_element

    return table_operation


def create_one_input_user_defined_data_stream_function_from_protos(
        function_infos, input_coder_info, output_coder_info, runtime_context,
        function_context, timer_context, side_output_context, job_parameters, keyed_state_backend,
        operator_state_backend):
    serialized_fns = [pare_user_defined_data_stream_function_proto(proto)
                      for proto in function_infos]
    input_data_converter = (
        from_type_info_proto(parse_coder_proto(input_coder_info).raw_type.type_info))
    output_data_converter = (
        from_type_info_proto(parse_coder_proto(output_coder_info).raw_type.type_info))

    function_operation = OneInputFunctionOperation(
        serialized_fns,
        input_data_converter,
        output_data_converter,
        runtime_context,
        function_context,
        timer_context,
        side_output_context,
        job_parameters,
        keyed_state_backend,
        operator_state_backend)

    return function_operation


def create_two_input_user_defined_data_stream_function_from_protos(
        function_infos, input_coder_info1, input_coder_info2, output_coder_info, runtime_context,
        function_context, timer_context, side_output_context, job_parameters, keyed_state_backend,
        operator_state_backend):
    serialized_fns = [pare_user_defined_data_stream_function_proto(proto)
                      for proto in function_infos]

    input_data_converter1 = (
        from_type_info_proto(parse_coder_proto(input_coder_info1).raw_type.type_info))

    input_data_converter2 = (
        from_type_info_proto(parse_coder_proto(input_coder_info2).raw_type.type_info))

    output_data_converter = (
        from_type_info_proto(parse_coder_proto(output_coder_info).raw_type.type_info))

    function_operation = TwoInputFunctionOperation(
        serialized_fns,
        input_data_converter1,
        input_data_converter2,
        output_data_converter,
        runtime_context,
        function_context,
        timer_context,
        side_output_context,
        job_parameters,
        keyed_state_backend,
        operator_state_backend)

    return function_operation



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

__all__ = ['MemorySize', 'SlotSharingGroup']

from typing import Optional

from pyflink.java_gateway import get_gateway


class MemorySize(object):
    """
    MemorySize is a representation of a number of bytes, viewable in different units.
    """

    def __init__(self, j_memory_size=None, bytes_size: int = None):
        self._j_memory_size = get_gateway().jvm \
            .org.apache.flink.configuration.MemorySize(bytes_size) \
            if j_memory_size is None else j_memory_size

    @staticmethod
    def of_mebi_bytes(mebi_bytes: int) -> 'MemorySize':
        return MemorySize(
            get_gateway().jvm.org.apache.flink.configuration.MemorySize.ofMebiBytes(mebi_bytes))

    def get_bytes(self) -> int:
        """
        Gets the memory size in bytes.

        :return: The memory size in bytes.
        """
        return self._j_memory_size.getBytes()

    def get_kibi_bytes(self) -> int:
        """
        Gets the memory size in Kibibytes (= 1024 bytes).

        :return: The memory size in Kibibytes.
        """
        return self._j_memory_size.getKibiBytes()

    def get_mebi_bytes(self) -> int:
        """
        Gets the memory size in Mebibytes (= 1024 Kibibytes).

        :return: The memory size in Mebibytes.
        """
        return self._j_memory_size.getMebiBytes()

    def get_gibi_bytes(self) -> int:
        """
        Gets the memory size in Gibibytes (= 1024 Mebibytes).

        :return: The memory size in Gibibytes.
        """
        return self._j_memory_size.getGibiBytes()

    def get_tebi_bytes(self) -> int:
        """
        Gets the memory size in Tebibytes (= 1024 Gibibytes).

        :return: The memory size in Tebibytes.
        """
        return self._j_memory_size.getTebiBytes()

    def get_java_memory_size(self):
        """
        Gets the Java MemorySize object.

        :return: The Java MemorySize object.
        """
        return self._j_memory_size

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._j_memory_size == other._j_memory_size

    def __hash__(self):
        return self._j_memory_size.hashCode()

    def __lt__(self, other: 'MemorySize'):
        if not isinstance(other, MemorySize):
            raise Exception("Does not support comparison with non-MemorySize %s" % other)

        return self._j_memory_size.compareTo(other._j_memory_size) == -1

    def __le__(self, other: 'MemorySize'):
        return self.__eq__(other) and self.__lt__(other)

    def __str__(self):
        return self._j_memory_size.toString()


class SlotSharingGroup(object):
    """
    Describe the name and the different resource components of a slot sharing group.
    """

    def __init__(self, j_slot_sharing_group):
        self._j_slot_sharing_group = j_slot_sharing_group

    def get_name(self) -> str:
        """
        Gets the name of this SlotSharingGroup.

        :return: The name of the SlotSharingGroup.
        """
        return self._j_slot_sharing_group.getName()

    def get_managed_memory(self) -> Optional[MemorySize]:
        """
        Gets the task managed memory for this SlotSharingGroup.

        :return: The task managed memory of the SlotSharingGroup.
        """
        managed_memory = self._j_slot_sharing_group.getManagedMemory()
        return MemorySize(managed_memory.get()) if managed_memory.isPresent() else None

    def get_task_heap_memory(self) -> Optional[MemorySize]:
        """
        Gets the task heap memory for this SlotSharingGroup.

        :return: The task heap memory of the SlotSharingGroup.
        """
        task_heap_memory = self._j_slot_sharing_group.getTaskHeapMemory()
        return MemorySize(task_heap_memory.get()) if task_heap_memory.isPresent() else None

    def get_task_off_heap_memory(self) -> Optional[MemorySize]:
        """
        Gets the task off-heap memory for this SlotSharingGroup.

        :return: The task off-heap memory of the SlotSharingGroup.
        """
        task_off_heap_memory = self._j_slot_sharing_group.getTaskOffHeapMemory()
        return MemorySize(task_off_heap_memory.get()) if task_off_heap_memory.isPresent() else None

    def get_cpu_cores(self) -> Optional[float]:
        """
       Gets the CPU cores for this SlotSharingGroup.

        :return: The CPU cores of the SlotSharingGroup.
        """
        cpu_cores = self._j_slot_sharing_group.getCpuCores()
        return cpu_cores.get() if cpu_cores.isPresent() else None

    def get_external_resources(self) -> dict:
        """
        Gets the external resource from this SlotSharingGroup.

        :return: User specified resources of the SlotSharingGroup.
        """
        return dict(self._j_slot_sharing_group.getExternalResources())

    def get_java_slot_sharing_group(self):
        """
        Gets the Java SlotSharingGroup object.

        :return: The Java SlotSharingGroup object.
        """
        return self._j_slot_sharing_group

    @staticmethod
    def builder(name: str) -> 'Builder':
        """
        Gets the Builder with the given name for this SlotSharingGroup.

        :param name: The name of the SlotSharingGroup.
        :return: The builder for the SlotSharingGroup.
        """
        return SlotSharingGroup.Builder(
            get_gateway().jvm.org.apache.flink.api.common.operators.SlotSharingGroup.newBuilder(
                name))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            self._j_slot_sharing_group == other._j_slot_sharing_group

    def __hash__(self):
        return self._j_slot_sharing_group.hashCode()

    class Builder(object):
        """
        Builder for the SlotSharingGroup.
        """

        def __init__(self, j_builder):
            self._j_builder = j_builder

        def set_cpu_cores(self, cpu_cores: float) -> 'SlotSharingGroup.Builder':
            """
            Sets the CPU cores for this SlotSharingGroup.

            :param cpu_cores: The CPU cores of the SlotSharingGroup.
            :return: This object.
            """
            self._j_builder.setCpuCores(cpu_cores)
            return self

        def set_task_heap_memory(self, task_heap_memory: MemorySize) -> 'SlotSharingGroup.Builder':
            """
            Sets the task heap memory for this SlotSharingGroup.

            :param task_heap_memory: The task heap memory of the SlotSharingGroup.
            :return: This object.
            """
            self._j_builder.setTaskHeapMemory(task_heap_memory.get_java_memory_size())
            return self

        def set_task_heap_memory_mb(self, task_heap_memory_mb: int) -> 'SlotSharingGroup.Builder':
            """
            Sets the task heap memory for this SlotSharingGroup in MB.

            :param task_heap_memory_mb: The task heap memory of the SlotSharingGroup in MB.
            :return: This object.
            """
            self._j_builder.setTaskHeapMemoryMB(task_heap_memory_mb)
            return self

        def set_task_off_heap_memory(self, task_off_heap_memory: MemorySize) \
                -> 'SlotSharingGroup.Builder':
            """
            Sets the task off-heap memory for this SlotSharingGroup.

            :param task_off_heap_memory: The task off-heap memory of the SlotSharingGroup.
            :return: This object.
            """
            self._j_builder.setTaskOffHeapMemory(task_off_heap_memory.get_java_memory_size())
            return self

        def set_task_off_heap_memory_mb(self, task_off_heap_memory_mb: int) \
                -> 'SlotSharingGroup.Builder':
            """
            Sets the task off-heap memory for this SlotSharingGroup in MB.

            :param task_off_heap_memory_mb: The task off-heap memory of the SlotSharingGroup in MB.
            :return: This object.
            """
            self._j_builder.setTaskOffHeapMemoryMB(task_off_heap_memory_mb)
            return self

        def set_managed_memory(self, managed_memory: MemorySize) -> 'SlotSharingGroup.Builder':
            """
            Sets the task managed memory for this SlotSharingGroup.

            :param managed_memory: The task managed memory of the SlotSharingGroup.
            :return: This object.
            """
            self._j_builder.setManagedMemory(managed_memory.get_java_memory_size())
            return self

        def set_managed_memory_mb(self, managed_memory_mb: int) -> 'SlotSharingGroup.Builder':
            """
            Sets the task managed memory for this SlotSharingGroup in MB.

            :param managed_memory_mb: The task managed memory of the SlotSharingGroup in MB.
            :return: This object.
            """
            self._j_builder.setManagedMemoryMB(managed_memory_mb)
            return self

        def set_external_resource(self, name: str, value: float) -> 'SlotSharingGroup.Builder':
            """
            Adds the given external resource. The old value with the same resource name will be
            replaced if present.

            :param name: The resource name of the given external resource.
            :param value: The value of the given external resource.
            :return: This object.
            """
            self._j_builder.setExternalResource(name, value)
            return self

        def build(self) -> 'SlotSharingGroup':
            """
            Builds the SlotSharingGroup.

            :return: The SlotSharingGroup object.
            """
            return SlotSharingGroup(j_slot_sharing_group=self._j_builder.build())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import datetime
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, List, Dict

import pytz

from pyflink.common import Row, RowKind
from pyflink.common.constants import MAX_LONG_VALUE
from pyflink.fn_execution.datastream.timerservice import InternalTimer
from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl
from pyflink.fn_execution.coders import PickleCoder
from pyflink.fn_execution.table.aggregate_slow import DistinctViewDescriptor, RowKeySelector
from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \
    PerWindowStateDataViewStore
from pyflink.fn_execution.table.window_assigner import WindowAssigner, PanedWindowAssigner, \
    MergingWindowAssigner
from pyflink.fn_execution.table.window_context import WindowContext, TriggerContext, K, W
from pyflink.fn_execution.table.window_process_function import GeneralWindowProcessFunction, \
    InternalWindowProcessFunction, PanedWindowProcessFunction, MergingWindowProcessFunction
from pyflink.fn_execution.table.window_trigger import Trigger
from pyflink.table.udf import ImperativeAggregateFunction, FunctionContext


N = TypeVar('N')


def join_row(left: List, right: List):
    return Row(*(left + right))


class NamespaceAggsHandleFunctionBase(Generic[N], ABC):
    @abstractmethod
    def open(self, state_data_view_store):
        """
        Initialization method for the function. It is called before the actual working methods.

        :param state_data_view_store: The object used to manage the DataView.
        """
        pass

    @abstractmethod
    def accumulate(self, input_data: Row):
        """
        Accumulates the input values to the accumulators.

        :param input_data: Input values bundled in a Row.
        """
        pass

    @abstractmethod
    def retract(self, input_data: Row):
        """
        Retracts the input values from the accumulators.

        :param input_data: Input values bundled in a Row.
        """

    @abstractmethod
    def merge(self, namespace: N, accumulators: List):
        """
        Merges the other accumulators into current accumulators.
        """
        pass

    @abstractmethod
    def set_accumulators(self, namespace: N, accumulators: List):
        """
        Set the current accumulators (saved in a row) which contains the current aggregated results.
        """
        pass

    @abstractmethod
    def get_accumulators(self) -> List:
        """
        Gets the current accumulators (saved in a list) which contains the current
        aggregated results.

        :return: The current accumulators.
        """
        pass

    @abstractmethod
    def create_accumulators(self) -> List:
        """
        Initializes the accumulators and save them to an accumulators List.

        :return: A List of accumulators which contains the aggregated results.
        """
        pass

    @abstractmethod
    def cleanup(self, namespace: N):
        """
        Cleanup for the retired accumulators state.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Tear-down method for this function. It can be used for clean up work.
        By default, this method does nothing.
        """
        pass


class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase[N], ABC):
    @abstractmethod
    def get_value(self, namespace: N) -> List:
        """
        Gets the result of the aggregation from the current accumulators and namespace properties
        (like window start).
        :param namespace: the namespace properties which should be calculated, such window start
        :return: the final result (saved in a List) of the current accumulators.
        """
        pass


class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction[N]):

    def __init__(self,
                 udfs: List[ImperativeAggregateFunction],
                 input_extractors: List,
                 index_of_count_star: int,
                 count_star_inserted: bool,
                 named_property_extractor,
                 udf_data_view_specs: List[List[DataViewSpec]],
                 filter_args: List[int],
                 distinct_indexes: List[int],
                 distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
        self._udfs = udfs
        self._input_extractors = input_extractors
        self._named_property_extractor = named_property_extractor
        self._accumulators = None  # type: List
        self._udf_data_view_specs = udf_data_view_specs
        self._udf_data_views = []
        self._filter_args = filter_args
        self._distinct_indexes = distinct_indexes
        self._distinct_view_descriptors = distinct_view_descriptors
        self._distinct_data_views = {}
        self._get_value_indexes = [i for i in range(len(udfs))]
        if index_of_count_star >= 0 and count_star_inserted:
            # The record count is used internally, should be ignored by the get_value method.
            self._get_value_indexes.remove(index_of_count_star)

    def open(self, state_data_view_store):
        for udf in self._udfs:
            udf.open(state_data_view_store.get_runtime_context())
        self._udf_data_views = []
        for data_view_specs in self._udf_data_view_specs:
            data_views = {}
            for data_view_spec in data_view_specs:
                if isinstance(data_view_spec, ListViewSpec):
                    data_views[data_view_spec.field_index] = \
                        state_data_view_store.get_state_list_view(
                            data_view_spec.state_id,
                            data_view_spec.element_coder)
                elif isinstance(data_view_spec, MapViewSpec):
                    data_views[data_view_spec.field_index] = \
                        state_data_view_store.get_state_map_view(
                            data_view_spec.state_id,
                            data_view_spec.key_coder,
                            data_view_spec.value_coder)
            self._udf_data_views.append(data_views)
        for key in self._distinct_view_descriptors.keys():
            self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
                "agg%ddistinct" % key,
                PickleCoder(),
                PickleCoder())

    def accumulate(self, input_data: Row):
        for i in range(len(self._udfs)):
            if i in self._distinct_data_views:
                if len(self._distinct_view_descriptors[i].get_filter_args()) == 0:
                    filtered = False
                else:
                    filtered = True
                    for filter_arg in self._distinct_view_descriptors[i].get_filter_args():
                        if input_data[filter_arg]:
                            filtered = False
                            break
                if not filtered:
                    input_extractor = self._distinct_view_descriptors[i].get_input_extractor()
                    args = input_extractor(input_data)
                    if args in self._distinct_data_views[i]:
                        self._distinct_data_views[i][args] += 1
                    else:
                        self._distinct_data_views[i][args] = 1
            if self._filter_args[i] >= 0 and not input_data[self._filter_args[i]]:
                continue
            input_extractor = self._input_extractors[i]
            args = input_extractor(input_data)
            if self._distinct_indexes[i] >= 0:
                if args in self._distinct_data_views[self._distinct_indexes[i]]:
                    if self._distinct_data_views[self._distinct_indexes[i]][args] > 1:
                        continue
                else:
                    raise Exception(
                        "The args are not in the distinct data view, this should not happen.")
            self._udfs[i].accumulate(self._accumulators[i], *args)

    def retract(self, input_data: Row):
        for i in range(len(self._udfs)):
            if i in self._distinct_data_views:
                if len(self._distinct_view_descriptors[i].get_filter_args()) == 0:
                    filtered = False
                else:
                    filtered = True
                    for filter_arg in self._distinct_view_descriptors[i].get_filter_args():
                        if input_data[filter_arg]:
                            filtered = False
                            break
                if not filtered:
                    input_extractor = self._distinct_view_descriptors[i].get_input_extractor()
                    args = input_extractor(input_data)
                    if args in self._distinct_data_views[i]:
                        self._distinct_data_views[i][args] -= 1
                        if self._distinct_data_views[i][args] == 0:
                            del self._distinct_data_views[i][args]
            if self._filter_args[i] >= 0 and not input_data[self._filter_args[i]]:
                continue
            input_extractor = self._input_extractors[i]
            args = input_extractor(input_data)
            if self._distinct_indexes[i] >= 0 and \
                    args in self._distinct_data_views[self._distinct_indexes[i]]:
                continue
            self._udfs[i].retract(self._accumulators[i], *args)

    def merge(self, namespace: N, accumulators: List):
        if self._udf_data_views:
            for i in range(len(self._udf_data_views)):
                for index, data_view in self._udf_data_views[i].items():
                    data_view.set_current_namespace(namespace)
                    accumulators[i][index] = data_view
        for i in range(len(self._udfs)):
            self._udfs[i].merge(self._accumulators[i], [accumulators[i]])

    def set_accumulators(self, namespace: N, accumulators: List):
        if self._udf_data_views and namespace is not None:
            for i in range(len(self._udf_data_views)):
                for index, data_view in self._udf_data_views[i].items():
                    data_view.set_current_namespace(namespace)
                    accumulators[i][index] = data_view
        self._accumulators = accumulators

    def get_accumulators(self) -> List:
        return self._accumulators

    def create_accumulators(self) -> List:
        return [udf.create_accumulator() for udf in self._udfs]

    def cleanup(self, namespace: N):
        for i in range(len(self._udf_data_views)):
            for data_view in self._udf_data_views[i].values():
                data_view.set_current_namespace(namespace)
                data_view.clear()

    def close(self):
        for udf in self._udfs:
            udf.close()

    def get_value(self, namespace: N) -> List:
        result = [self._udfs[i].get_value(self._accumulators[i]) for i in self._get_value_indexes]
        if self._named_property_extractor:
            result.extend(self._named_property_extractor(namespace))
        return result


class GroupWindowAggFunctionBase(Generic[K, W]):
    def __init__(self,
                 allowed_lateness: int,
                 key_selector: RowKeySelector,
                 state_backend,
                 state_value_coder,
                 window_assigner: WindowAssigner[W],
                 window_aggregator: NamespaceAggsHandleFunctionBase[W],
                 trigger: Trigger[W],
                 rowtime_index: int,
                 shift_timezone: str):
        self._allowed_lateness = allowed_lateness
        self._key_selector = key_selector
        self._state_backend = state_backend
        self._state_value_coder = state_value_coder
        self._window_assigner = window_assigner
        self._window_aggregator = window_aggregator
        self._rowtime_index = rowtime_index
        self._shift_timezone = shift_timezone
        self._window_function = None  # type: InternalWindowProcessFunction[K, W]
        self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl
        self._window_context = None  # type: WindowContext
        self._trigger = trigger
        self._trigger_context = None  # type: TriggerContext
        self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)

    def open(self, function_context: FunctionContext):
        self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
        self._window_aggregator.open(
            PerWindowStateDataViewStore(function_context, self._state_backend))

        if isinstance(self._window_assigner, PanedWindowAssigner):
            self._window_function = PanedWindowProcessFunction(
                self._allowed_lateness, self._window_assigner, self._window_aggregator)
        elif isinstance(self._window_assigner, MergingWindowAssigner):
            self._window_function = MergingWindowProcessFunction(
                self._allowed_lateness, self._window_assigner, self._window_aggregator,
                self._state_backend)
        else:
            self._window_function = GeneralWindowProcessFunction(
                self._allowed_lateness, self._window_assigner, self._window_aggregator)
        self._trigger_context = TriggerContext(
            self._trigger, self._internal_timer_service, self._state_backend)
        self._trigger_context.open()
        self._window_context = WindowContext(
            self, self._trigger_context, self._state_backend, self._state_value_coder,
            self._internal_timer_service, self._window_assigner.is_event_time())
        self._window_function.open(self._window_context)

    def process_element(self, input_row: Row):
        input_value = input_row._values
        current_key = self._key_selector.get_key(input_value)
        self._state_backend.set_current_key(current_key)
        if self._window_assigner.is_event_time():
            timestamp = input_value[self._rowtime_index]
            seconds = int(timestamp.replace(tzinfo=datetime.timezone.utc).timestamp())
            microseconds_of_second = timestamp.microsecond
            milliseconds = seconds * 1000 + microseconds_of_second // 1000
            timestamp = milliseconds
        else:
            timestamp = self._internal_timer_service.current_processing_time()

        timestamp = self.to_utc_timestamp_mills(timestamp)

        # the windows which the input row should be placed into
        affected_windows = self._window_function.assign_state_namespace(input_value, timestamp)
        for window in affected_windows:
            self._window_state.set_current_namespace(window)
            acc = self._window_state.value()  # type: List
            if acc is None:
                acc = self._window_aggregator.create_accumulators()
            self._window_aggregator.set_accumulators(window, acc)

            if input_row._is_accumulate_msg():
                self._window_aggregator.accumulate(input_row)
            else:
                self._window_aggregator.retract(input_row)
            acc = self._window_aggregator.get_accumulators()
            self._window_state.update(acc)

        # the actual window which the input row is belongs to
        actual_windows = self._window_function.assign_actual_windows(input_value, timestamp)
        result = []
        for window in actual_windows:
            self._trigger_context.window = window
            trigger_result = self._trigger_context.on_element(input_row, timestamp)
            if trigger_result:
                result.append(self._emit_window_result(current_key, window))
            self._register_cleanup_timer(window)
        return result

    def process_watermark(self, watermark: int):
        self._internal_timer_service.advance_watermark(watermark)

    def on_event_time(self, timer: InternalTimer):
        result = []
        timestamp = timer.get_timestamp()
        key = timer.get_key()
        self._state_backend.set_current_key(key)
        window = timer.get_namespace()
        self._trigger_context.window = window
        if self._trigger_context.on_event_time(timestamp):
            # fire
            result.append(self._emit_window_result(key, window))

        if self._window_assigner.is_event_time():
            self._window_function.clean_window_if_needed(window, timestamp)
        return result

    def on_processing_time(self, timer: InternalTimer):
        result = []
        timestamp = timer.get_timestamp()
        key = timer.get_key()
        self._state_backend.set_current_key(key)
        window = timer.get_namespace()
        self._trigger_context.window = window
        if self._trigger_context.on_processing_time(timestamp):
            # fire
            result.append(self._emit_window_result(key, window))

        if not self._window_assigner.is_event_time():
            self._window_function.clean_window_if_needed(window, timestamp)
        return result

    def get_timers(self):
        yield from self._internal_timer_service._timers.keys()
        self._internal_timer_service._timers.clear()

    def to_utc_timestamp_mills(self, epoch_mills):
        if self._shift_timezone == "UTC":
            return epoch_mills
        else:
            timezone = pytz.timezone(self._shift_timezone)
            local_date_time = datetime.datetime.fromtimestamp(epoch_mills / 1000., timezone)\
                .replace(tzinfo=None)
            epoch = datetime.datetime.utcfromtimestamp(0)
            return int((local_date_time - epoch).total_seconds() * 1000.0)

    def close(self):
        self._window_aggregator.close()

    def _register_cleanup_timer(self, window: N):
        cleanup_time = self.cleanup_time(window)
        if cleanup_time == MAX_LONG_VALUE:
            return

        if self._window_assigner.is_event_time():
            self._trigger_context.register_event_time_timer(cleanup_time)
        else:
            self._trigger_context.register_processing_time_timer(cleanup_time)

    def cleanup_time(self, window: N) -> int:
        if self._window_assigner.is_event_time():
            cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
            if cleanup_time >= window.max_timestamp():
                return cleanup_time
            else:
                return MAX_LONG_VALUE
        else:
            return max(0, window.max_timestamp())

    @abstractmethod
    def _emit_window_result(self, key: List, window: W):
        pass


class GroupWindowAggFunction(GroupWindowAggFunctionBase[K, W]):

    def __init__(self,
                 allowed_lateness: int,
                 key_selector: RowKeySelector,
                 state_backend,
                 state_value_coder,
                 window_assigner: WindowAssigner[W],
                 window_aggregator: NamespaceAggsHandleFunction[W],
                 trigger: Trigger[W],
                 rowtime_index: int,
                 shift_timezone: str):
        super(GroupWindowAggFunction, self).__init__(
            allowed_lateness, key_selector, state_backend, state_value_coder, window_assigner,
            window_aggregator, trigger, rowtime_index, shift_timezone)
        self._window_aggregator = window_aggregator

    def _emit_window_result(self, key: List, window: W):
        self._window_function.prepare_aggregate_accumulator_for_emit(window)
        agg_result = self._window_aggregator.get_value(window)
        result_row = join_row(key, agg_result)
        # send INSERT
        result_row.set_row_kind(RowKind.INSERT)
        return result_row



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################

from pyflink.java_gateway import get_gateway


def results():
    """
    Retrieves the results from an append table sink.
    """
    return retract_results()


def retract_results():
    """
    Retrieves the results from a retract table sink.
    """
    gateway = get_gateway()
    results = gateway.jvm.org.apache.flink.table.utils.TestingSinks.RowCollector.getAndClearValues()
    return gateway.jvm\
        .org.apache.flink.table.utils.TestingSinks.RowCollector.retractResults(results)


def upsert_results(keys):
    """
    Retrieves the results from an upsert table sink.
    """
    gateway = get_gateway()
    j_keys = gateway.new_array(gateway.jvm.int, len(keys))
    for i in range(0, len(keys)):
        j_keys[i] = keys[i]

    results = gateway.jvm.RowCollector.getAndClearValues()
    return gateway.jvm.RowCollector.upsertResults(results, j_keys)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: flink-fn-execution.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66link-fn-execution.proto\x12 org.apache.flink.fn_execution.v1\"*\n\x0cJobParameter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x86\x01\n\x05Input\x12\x44\n\x03udf\x18\x01 \x01(\x0b\x32\x35.org.apache.flink.fn_execution.v1.UserDefinedFunctionH\x00\x12\x15\n\x0binputOffset\x18\x02 \x01(\x05H\x00\x12\x17\n\rinputConstant\x18\x03 \x01(\x0cH\x00\x42\x07\n\x05input\"\xa8\x01\n\x13UserDefinedFunction\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x37\n\x06inputs\x18\x02 \x03(\x0b\x32\'.org.apache.flink.fn_execution.v1.Input\x12\x14\n\x0cwindow_index\x18\x03 \x01(\x05\x12\x1a\n\x12takes_row_as_input\x18\x04 \x01(\x08\x12\x15\n\ris_pandas_udf\x18\x05 \x01(\x08\"\x93\x02\n\x14UserDefinedFunctions\x12\x43\n\x04udfs\x18\x01 \x03(\x0b\x32\x35.org.apache.flink.fn_execution.v1.UserDefinedFunction\x12\x16\n\x0emetric_enabled\x18\x02 \x01(\x08\x12=\n\x07windows\x18\x03 \x03(\x0b\x32,.org.apache.flink.fn_execution.v1.OverWindow\x12\x17\n\x0fprofile_enabled\x18\x04 \x01(\x08\x12\x46\n\x0ejob_parameters\x18\x05 \x03(\x0b\x32..org.apache.flink.fn_execution.v1.JobParameter\"\xdd\x02\n\nOverWindow\x12L\n\x0bwindow_type\x18\x01 \x01(\x0e\x32\x37.org.apache.flink.fn_execution.v1.OverWindow.WindowType\x12\x16\n\x0elower_boundary\x18\x02 \x01(\x03\x12\x16\n\x0eupper_boundary\x18\x03 \x01(\x03\"\xd0\x01\n\nWindowType\x12\x13\n\x0fRANGE_UNBOUNDED\x10\x00\x12\x1d\n\x19RANGE_UNBOUNDED_PRECEDING\x10\x01\x12\x1d\n\x19RANGE_UNBOUNDED_FOLLOWING\x10\x02\x12\x11\n\rRANGE_SLIDING\x10\x03\x12\x11\n\rROW_UNBOUNDED\x10\x04\x12\x1b\n\x17ROW_UNBOUNDED_PRECEDING\x10\x05\x12\x1b\n\x17ROW_UNBOUNDED_FOLLOWING\x10\x06\x12\x0f\n\x0bROW_SLIDING\x10\x07\"\x8b\x06\n\x1cUserDefinedAggregateFunction\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x37\n\x06inputs\x18\x02 \x03(\x0b\x32\'.org.apache.flink.fn_execution.v1.Input\x12Z\n\x05specs\x18\x03 \x03(\x0b\x32K.org.apache.flink.fn_execution.v1.UserDefinedAggregateFunction.DataViewSpec\x12\x12\n\nfilter_arg\x18\x04 \x01(\x05\x12\x10\n\x08\x64istinct\x18\x05 \x01(\x08\x12\x1a\n\x12takes_row_as_input\x18\x06 \x01(\x08\x1a\x82\x04\n\x0c\x44\x61taViewSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x66ield_index\x18\x02 \x01(\x05\x12i\n\tlist_view\x18\x03 \x01(\x0b\x32T.org.apache.flink.fn_execution.v1.UserDefinedAggregateFunction.DataViewSpec.ListViewH\x00\x12g\n\x08map_view\x18\x04 \x01(\x0b\x32S.org.apache.flink.fn_execution.v1.UserDefinedAggregateFunction.DataViewSpec.MapViewH\x00\x1aT\n\x08ListView\x12H\n\x0c\x65lement_type\x18\x01 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\x1a\x97\x01\n\x07MapView\x12\x44\n\x08key_type\x18\x01 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\x12\x46\n\nvalue_type\x18\x02 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldTypeB\x0b\n\tdata_view\"\xac\x04\n\x0bGroupWindow\x12M\n\x0bwindow_type\x18\x01 \x01(\x0e\x32\x38.org.apache.flink.fn_execution.v1.GroupWindow.WindowType\x12\x16\n\x0eis_time_window\x18\x02 \x01(\x08\x12\x14\n\x0cwindow_slide\x18\x03 \x01(\x03\x12\x13\n\x0bwindow_size\x18\x04 \x01(\x03\x12\x12\n\nwindow_gap\x18\x05 \x01(\x03\x12\x13\n\x0bis_row_time\x18\x06 \x01(\x08\x12\x18\n\x10time_field_index\x18\x07 \x01(\x05\x12\x17\n\x0f\x61llowedLateness\x18\x08 \x01(\x03\x12U\n\x0fnamedProperties\x18\t \x03(\x0e\x32<.org.apache.flink.fn_execution.v1.GroupWindow.WindowProperty\x12\x16\n\x0eshift_timezone\x18\n \x01(\t\"[\n\nWindowType\x12\x19\n\x15TUMBLING_GROUP_WINDOW\x10\x00\x12\x18\n\x14SLIDING_GROUP_WINDOW\x10\x01\x12\x18\n\x14SESSION_GROUP_WINDOW\x10\x02\"c\n\x0eWindowProperty\x12\x10\n\x0cWINDOW_START\x10\x00\x12\x0e\n\nWINDOW_END\x10\x01\x12\x16\n\x12ROW_TIME_ATTRIBUTE\x10\x02\x12\x17\n\x13PROC_TIME_ATTRIBUTE\x10\x03\"\xde\x04\n\x1dUserDefinedAggregateFunctions\x12L\n\x04udfs\x18\x01 \x03(\x0b\x32>.org.apache.flink.fn_execution.v1.UserDefinedAggregateFunction\x12\x16\n\x0emetric_enabled\x18\x02 \x01(\x08\x12\x10\n\x08grouping\x18\x03 \x03(\x05\x12\x1e\n\x16generate_update_before\x18\x04 \x01(\x08\x12\x44\n\x08key_type\x18\x05 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\x12\x1b\n\x13index_of_count_star\x18\x06 \x01(\x05\x12\x1e\n\x16state_cleaning_enabled\x18\x07 \x01(\x08\x12\x18\n\x10state_cache_size\x18\x08 \x01(\x05\x12!\n\x19map_state_read_cache_size\x18\t \x01(\x05\x12\"\n\x1amap_state_write_cache_size\x18\n \x01(\x05\x12\x1b\n\x13\x63ount_star_inserted\x18\x0b \x01(\x08\x12\x43\n\x0cgroup_window\x18\x0c \x01(\x0b\x32-.org.apache.flink.fn_execution.v1.GroupWindow\x12\x17\n\x0fprofile_enabled\x18\r \x01(\x08\x12\x46\n\x0ejob_parameters\x18\x0e \x03(\x0b\x32..org.apache.flink.fn_execution.v1.JobParameter\"\xf6\x0f\n\x06Schema\x12>\n\x06\x66ields\x18\x01 \x03(\x0b\x32..org.apache.flink.fn_execution.v1.Schema.Field\x1a\x97\x01\n\x07MapInfo\x12\x44\n\x08key_type\x18\x01 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\x12\x46\n\nvalue_type\x18\x02 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\x1a\x1d\n\x08TimeInfo\x12\x11\n\tprecision\x18\x01 \x01(\x05\x1a\"\n\rTimestampInfo\x12\x11\n\tprecision\x18\x01 \x01(\x05\x1a,\n\x17LocalZonedTimestampInfo\x12\x11\n\tprecision\x18\x01 \x01(\x05\x1a\'\n\x12ZonedTimestampInfo\x12\x11\n\tprecision\x18\x01 \x01(\x05\x1a/\n\x0b\x44\x65\x63imalInfo\x12\x11\n\tprecision\x18\x01 \x01(\x05\x12\r\n\x05scale\x18\x02 \x01(\x05\x1a\x1c\n\nBinaryInfo\x12\x0e\n\x06length\x18\x01 \x01(\x05\x1a\x1f\n\rVarBinaryInfo\x12\x0e\n\x06length\x18\x01 \x01(\x05\x1a\x1a\n\x08\x43harInfo\x12\x0e\n\x06length\x18\x01 \x01(\x05\x1a\x1d\n\x0bVarCharInfo\x12\x0e\n\x06length\x18\x01 \x01(\x05\x1a\xb0\x08\n\tFieldType\x12\x44\n\ttype_name\x18\x01 \x01(\x0e\x32\x31.org.apache.flink.fn_execution.v1.Schema.TypeName\x12\x10\n\x08nullable\x18\x02 \x01(\x08\x12U\n\x17\x63ollection_element_type\x18\x03 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldTypeH\x00\x12\x44\n\x08map_info\x18\x04 \x01(\x0b\x32\x30.org.apache.flink.fn_execution.v1.Schema.MapInfoH\x00\x12>\n\nrow_schema\x18\x05 \x01(\x0b\x32(.org.apache.flink.fn_execution.v1.SchemaH\x00\x12L\n\x0c\x64\x65\x63imal_info\x18\x06 \x01(\x0b\x32\x34.org.apache.flink.fn_execution.v1.Schema.DecimalInfoH\x00\x12\x46\n\ttime_info\x18\x07 \x01(\x0b\x32\x31.org.apache.flink.fn_execution.v1.Schema.TimeInfoH\x00\x12P\n\x0etimestamp_info\x18\x08 \x01(\x0b\x32\x36.org.apache.flink.fn_execution.v1.Schema.TimestampInfoH\x00\x12\x66\n\x1alocal_zoned_timestamp_info\x18\t \x01(\x0b\x32@.org.apache.flink.fn_execution.v1.Schema.LocalZonedTimestampInfoH\x00\x12[\n\x14zoned_timestamp_info\x18\n \x01(\x0b\x32;.org.apache.flink.fn_execution.v1.Schema.ZonedTimestampInfoH\x00\x12J\n\x0b\x62inary_info\x18\x0b \x01(\x0b\x32\x33.org.apache.flink.fn_execution.v1.Schema.BinaryInfoH\x00\x12Q\n\x0fvar_binary_info\x18\x0c \x01(\x0b\x32\x36.org.apache.flink.fn_execution.v1.Schema.VarBinaryInfoH\x00\x12\x46\n\tchar_info\x18\r \x01(\x0b\x32\x31.org.apache.flink.fn_execution.v1.Schema.CharInfoH\x00\x12M\n\rvar_char_info\x18\x0e \x01(\x0b\x32\x34.org.apache.flink.fn_execution.v1.Schema.VarCharInfoH\x00\x42\x0b\n\ttype_info\x1al\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12@\n\x04type\x18\x03 \x01(\x0b\x32\x32.org.apache.flink.fn_execution.v1.Schema.FieldType\"\xab\x02\n\x08TypeName\x12\x07\n\x03ROW\x10\x00\x12\x0b\n\x07TINYINT\x10\x01\x12\x0c\n\x08SMALLINT\x10\x02\x12\x07\n\x03INT\x10\x03\x12\n\n\x06\x42IGINT\x10\x04\x12\x0b\n\x07\x44\x45\x43IMAL\x10\x05\x12\t\n\x05\x46LOAT\x10\x06\x12\n\n\x06\x44OUBLE\x10\x07\x12\x08\n\x04\x44\x41TE\x10\x08\x12\x08\n\x04TIME\x10\t\x12\r\n\tTIMESTAMP\x10\n\x12\x0b\n\x07\x42OOLEAN\x10\x0b\x12\n\n\x06\x42INARY\x10\x0c\x12\r\n\tVARBINARY\x10\r\x12\x08\n\x04\x43HAR\x10\x0e\x12\x0b\n\x07VARCHAR\x10\x0f\x12\x0f\n\x0b\x42\x41SIC_ARRAY\x10\x10\x12\x07\n\x03MAP\x10\x11\x12\x0c\n\x08MULTISET\x10\x12\x12\x19\n\x15LOCAL_ZONED_TIMESTAMP\x10\x13\x12\x13\n\x0fZONED_TIMESTAMP\x10\x14\x12\x08\n\x04NULL\x10\x15\"\xc3\n\n\x08TypeInfo\x12\x46\n\ttype_name\x18\x01 \x01(\x0e\x32\x33.org.apache.flink.fn_execution.v1.TypeInfo.TypeName\x12M\n\x17\x63ollection_element_type\x18\x02 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfoH\x00\x12O\n\rrow_type_info\x18\x03 \x01(\x0b\x32\x36.org.apache.flink.fn_execution.v1.TypeInfo.RowTypeInfoH\x00\x12S\n\x0ftuple_type_info\x18\x04 \x01(\x0b\x32\x38.org.apache.flink.fn_execution.v1.TypeInfo.TupleTypeInfoH\x00\x12O\n\rmap_type_info\x18\x05 \x01(\x0b\x32\x36.org.apache.flink.fn_execution.v1.TypeInfo.MapTypeInfoH\x00\x12Q\n\x0e\x61vro_type_info\x18\x06 \x01(\x0b\x32\x37.org.apache.flink.fn_execution.v1.TypeInfo.AvroTypeInfoH\x00\x1a\x8b\x01\n\x0bMapTypeInfo\x12<\n\x08key_type\x18\x01 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\x12>\n\nvalue_type\x18\x02 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\x1a\xb8\x01\n\x0bRowTypeInfo\x12L\n\x06\x66ields\x18\x01 \x03(\x0b\x32<.org.apache.flink.fn_execution.v1.TypeInfo.RowTypeInfo.Field\x1a[\n\x05\x46ield\x12\x12\n\nfield_name\x18\x01 \x01(\t\x12>\n\nfield_type\x18\x02 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\x1aP\n\rTupleTypeInfo\x12?\n\x0b\x66ield_types\x18\x01 \x03(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\x1a\x1e\n\x0c\x41vroTypeInfo\x12\x0e\n\x06schema\x18\x01 \x01(\t\"\x8d\x03\n\x08TypeName\x12\x07\n\x03ROW\x10\x00\x12\n\n\x06STRING\x10\x01\x12\x08\n\x04\x42YTE\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\t\n\x05SHORT\x10\x04\x12\x07\n\x03INT\x10\x05\x12\x08\n\x04LONG\x10\x06\x12\t\n\x05\x46LOAT\x10\x07\x12\n\n\x06\x44OUBLE\x10\x08\x12\x08\n\x04\x43HAR\x10\t\x12\x0b\n\x07\x42IG_INT\x10\n\x12\x0b\n\x07\x42IG_DEC\x10\x0b\x12\x0c\n\x08SQL_DATE\x10\x0c\x12\x0c\n\x08SQL_TIME\x10\r\x12\x11\n\rSQL_TIMESTAMP\x10\x0e\x12\x0f\n\x0b\x42\x41SIC_ARRAY\x10\x0f\x12\x13\n\x0fPRIMITIVE_ARRAY\x10\x10\x12\t\n\x05TUPLE\x10\x11\x12\x08\n\x04LIST\x10\x12\x12\x07\n\x03MAP\x10\x13\x12\x11\n\rPICKLED_BYTES\x10\x14\x12\x10\n\x0cOBJECT_ARRAY\x10\x15\x12\x0b\n\x07INSTANT\x10\x16\x12\x08\n\x04\x41VRO\x10\x17\x12\x0e\n\nLOCAL_DATE\x10\x18\x12\x0e\n\nLOCAL_TIME\x10\x19\x12\x12\n\x0eLOCAL_DATETIME\x10\x1a\x12\x19\n\x15LOCAL_ZONED_TIMESTAMP\x10\x1b\x42\x0b\n\ttype_info\"\xd1\x07\n\x1dUserDefinedDataStreamFunction\x12\x63\n\rfunction_type\x18\x01 \x01(\x0e\x32L.org.apache.flink.fn_execution.v1.UserDefinedDataStreamFunction.FunctionType\x12g\n\x0fruntime_context\x18\x02 \x01(\x0b\x32N.org.apache.flink.fn_execution.v1.UserDefinedDataStreamFunction.RuntimeContext\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x12\x16\n\x0emetric_enabled\x18\x04 \x01(\x08\x12\x41\n\rkey_type_info\x18\x05 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\x12\x17\n\x0fprofile_enabled\x18\x06 \x01(\x08\x12\x17\n\x0fhas_side_output\x18\x07 \x01(\x08\x12\x18\n\x10state_cache_size\x18\x08 \x01(\x05\x12!\n\x19map_state_read_cache_size\x18\t \x01(\x05\x12\"\n\x1amap_state_write_cache_size\x18\n \x01(\x05\x1a\xb2\x02\n\x0eRuntimeContext\x12\x11\n\ttask_name\x18\x01 \x01(\t\x12\x1f\n\x17task_name_with_subtasks\x18\x02 \x01(\t\x12#\n\x1bnumber_of_parallel_subtasks\x18\x03 \x01(\x05\x12\'\n\x1fmax_number_of_parallel_subtasks\x18\x04 \x01(\x05\x12\x1d\n\x15index_of_this_subtask\x18\x05 \x01(\x05\x12\x16\n\x0e\x61ttempt_number\x18\x06 \x01(\x05\x12\x46\n\x0ejob_parameters\x18\x07 \x03(\x0b\x32..org.apache.flink.fn_execution.v1.JobParameter\x12\x1f\n\x17in_batch_execution_mode\x18\x08 \x01(\x08\"\xad\x01\n\x0c\x46unctionType\x12\x0b\n\x07PROCESS\x10\x00\x12\x0e\n\nCO_PROCESS\x10\x01\x12\x11\n\rKEYED_PROCESS\x10\x02\x12\x14\n\x10KEYED_CO_PROCESS\x10\x03\x12\n\n\x06WINDOW\x10\x04\x12\x18\n\x14\x43O_BROADCAST_PROCESS\x10\x05\x12\x1e\n\x1aKEYED_CO_BROADCAST_PROCESS\x10\x06\x12\x11\n\rREVISE_OUTPUT\x10\x64\"\xe4\x0e\n\x0fStateDescriptor\x12\x12\n\nstate_name\x18\x01 \x01(\t\x12Z\n\x10state_ttl_config\x18\x02 \x01(\x0b\x32@.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig\x1a\xe0\r\n\x0eStateTTLConfig\x12`\n\x0bupdate_type\x18\x01 \x01(\x0e\x32K.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.UpdateType\x12j\n\x10state_visibility\x18\x02 \x01(\x0e\x32P.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.StateVisibility\x12w\n\x17ttl_time_characteristic\x18\x03 \x01(\x0e\x32V.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.TtlTimeCharacteristic\x12\x0b\n\x03ttl\x18\x04 \x01(\x03\x12n\n\x12\x63leanup_strategies\x18\x05 \x01(\x0b\x32R.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies\x1a\xca\x08\n\x11\x43leanupStrategies\x12 \n\x18is_cleanup_in_background\x18\x01 \x01(\x08\x12y\n\nstrategies\x18\x02 \x03(\x0b\x32\x65.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies.MapStrategiesEntry\x1aX\n\x1aIncrementalCleanupStrategy\x12\x14\n\x0c\x63leanup_size\x18\x01 \x01(\x05\x12$\n\x1crun_cleanup_for_every_record\x18\x02 \x01(\x08\x1aK\n#RocksdbCompactFilterCleanupStrategy\x12$\n\x1cquery_time_after_num_entries\x18\x01 \x01(\x03\x1a\xe0\x04\n\x12MapStrategiesEntry\x12o\n\x08strategy\x18\x01 \x01(\x0e\x32].org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies\x12\x81\x01\n\x0e\x65mpty_strategy\x18\x02 \x01(\x0e\x32g.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies.EmptyCleanupStrategyH\x00\x12\x95\x01\n\x1cincremental_cleanup_strategy\x18\x03 \x01(\x0b\x32m.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies.IncrementalCleanupStrategyH\x00\x12\xa9\x01\n\'rocksdb_compact_filter_cleanup_strategy\x18\x04 \x01(\x0b\x32v.org.apache.flink.fn_execution.v1.StateDescriptor.StateTTLConfig.CleanupStrategies.RocksdbCompactFilterCleanupStrategyH\x00\x42\x11\n\x0f\x43leanupStrategy\"b\n\nStrategies\x12\x1c\n\x18\x46ULL_STATE_SCAN_SNAPSHOT\x10\x00\x12\x17\n\x13INCREMENTAL_CLEANUP\x10\x01\x12\x1d\n\x19ROCKSDB_COMPACTION_FILTER\x10\x02\"*\n\x14\x45mptyCleanupStrategy\x12\x12\n\x0e\x45MPTY_STRATEGY\x10\x00\"D\n\nUpdateType\x12\x0c\n\x08\x44isabled\x10\x00\x12\x14\n\x10OnCreateAndWrite\x10\x01\x12\x12\n\x0eOnReadAndWrite\x10\x02\"J\n\x0fStateVisibility\x12\x1f\n\x1bReturnExpiredIfNotCleanedUp\x10\x00\x12\x16\n\x12NeverReturnExpired\x10\x01\"+\n\x15TtlTimeCharacteristic\x12\x12\n\x0eProcessingTime\x10\x00\"\xf1\x07\n\x13\x43oderInfoDescriptor\x12`\n\x10\x66latten_row_type\x18\x01 \x01(\x0b\x32\x44.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.FlattenRowTypeH\x00\x12Q\n\x08row_type\x18\x02 \x01(\x0b\x32=.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.RowTypeH\x00\x12U\n\narrow_type\x18\x03 \x01(\x0b\x32?.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.ArrowTypeH\x00\x12k\n\x16over_window_arrow_type\x18\x04 \x01(\x0b\x32I.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.OverWindowArrowTypeH\x00\x12Q\n\x08raw_type\x18\x05 \x01(\x0b\x32=.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.RawTypeH\x00\x12H\n\x04mode\x18\x06 \x01(\x0e\x32:.org.apache.flink.fn_execution.v1.CoderInfoDescriptor.Mode\x12\"\n\x1aseparated_with_end_message\x18\x07 \x01(\x08\x1aJ\n\x0e\x46lattenRowType\x12\x38\n\x06schema\x18\x01 \x01(\x0b\x32(.org.apache.flink.fn_execution.v1.Schema\x1a\x43\n\x07RowType\x12\x38\n\x06schema\x18\x01 \x01(\x0b\x32(.org.apache.flink.fn_execution.v1.Schema\x1a\x45\n\tArrowType\x12\x38\n\x06schema\x18\x01 \x01(\x0b\x32(.org.apache.flink.fn_execution.v1.Schema\x1aO\n\x13OverWindowArrowType\x12\x38\n\x06schema\x18\x01 \x01(\x0b\x32(.org.apache.flink.fn_execution.v1.Schema\x1aH\n\x07RawType\x12=\n\ttype_info\x18\x01 \x01(\x0b\x32*.org.apache.flink.fn_execution.v1.TypeInfo\" \n\x04Mode\x12\n\n\x06SINGLE\x10\x00\x12\x0c\n\x08MULTIPLE\x10\x01\x42\x0b\n\tdata_typeB-\n\x1forg.apache.flink.fnexecution.v1B\nFlinkFnApib\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'flink_fn_execution_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037org.apache.flink.fnexecution.v1B\nFlinkFnApi'
  _JOBPARAMETER._serialized_start=62
  _JOBPARAMETER._serialized_end=104
  _INPUT._serialized_start=107
  _INPUT._serialized_end=241
  _USERDEFINEDFUNCTION._serialized_start=244
  _USERDEFINEDFUNCTION._serialized_end=412
  _USERDEFINEDFUNCTIONS._serialized_start=415
  _USERDEFINEDFUNCTIONS._serialized_end=690
  _OVERWINDOW._serialized_start=693
  _OVERWINDOW._serialized_end=1042
  _OVERWINDOW_WINDOWTYPE._serialized_start=834
  _OVERWINDOW_WINDOWTYPE._serialized_end=1042
  _USERDEFINEDAGGREGATEFUNCTION._serialized_start=1045
  _USERDEFINEDAGGREGATEFUNCTION._serialized_end=1824
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC._serialized_start=1310
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC._serialized_end=1824
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC_LISTVIEW._serialized_start=1573
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC_LISTVIEW._serialized_end=1657
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC_MAPVIEW._serialized_start=1660
  _USERDEFINEDAGGREGATEFUNCTION_DATAVIEWSPEC_MAPVIEW._serialized_end=1811
  _GROUPWINDOW._serialized_start=1827
  _GROUPWINDOW._serialized_end=2383
  _GROUPWINDOW_WINDOWTYPE._serialized_start=2191
  _GROUPWINDOW_WINDOWTYPE._serialized_end=2282
  _GROUPWINDOW_WINDOWPROPERTY._serialized_start=2284
  _GROUPWINDOW_WINDOWPROPERTY._serialized_end=2383
  _USERDEFINEDAGGREGATEFUNCTIONS._serialized_start=2386
  _USERDEFINEDAGGREGATEFUNCTIONS._serialized_end=2992
  _SCHEMA._serialized_start=2995
  _SCHEMA._serialized_end=5033
  _SCHEMA_MAPINFO._serialized_start=3070
  _SCHEMA_MAPINFO._serialized_end=3221
  _SCHEMA_TIMEINFO._serialized_start=3223
  _SCHEMA_TIMEINFO._serialized_end=3252
  _SCHEMA_TIMESTAMPINFO._serialized_start=3254
  _SCHEMA_TIMESTAMPINFO._serialized_end=3288
  _SCHEMA_LOCALZONEDTIMESTAMPINFO._serialized_start=3290
  _SCHEMA_LOCALZONEDTIMESTAMPINFO._serialized_end=3334
  _SCHEMA_ZONEDTIMESTAMPINFO._serialized_start=3336
  _SCHEMA_ZONEDTIMESTAMPINFO._serialized_end=3375
  _SCHEMA_DECIMALINFO._serialized_start=3377
  _SCHEMA_DECIMALINFO._serialized_end=3424
  _SCHEMA_BINARYINFO._serialized_start=3426
  _SCHEMA_BINARYINFO._serialized_end=3454
  _SCHEMA_VARBINARYINFO._serialized_start=3456
  _SCHEMA_VARBINARYINFO._serialized_end=3487
  _SCHEMA_CHARINFO._serialized_start=3489
  _SCHEMA_CHARINFO._serialized_end=3515
  _SCHEMA_VARCHARINFO._serialized_start=3517
  _SCHEMA_VARCHARINFO._serialized_end=3546
  _SCHEMA_FIELDTYPE._serialized_start=3549
  _SCHEMA_FIELDTYPE._serialized_end=4621
  _SCHEMA_FIELD._serialized_start=4623
  _SCHEMA_FIELD._serialized_end=4731
  _SCHEMA_TYPENAME._serialized_start=4734
  _SCHEMA_TYPENAME._serialized_end=5033
  _TYPEINFO._serialized_start=5036
  _TYPEINFO._serialized_end=6383
  _TYPEINFO_MAPTYPEINFO._serialized_start=5530
  _TYPEINFO_MAPTYPEINFO._serialized_end=5669
  _TYPEINFO_ROWTYPEINFO._serialized_start=5672
  _TYPEINFO_ROWTYPEINFO._serialized_end=5856
  _TYPEINFO_ROWTYPEINFO_FIELD._serialized_start=5765
  _TYPEINFO_ROWTYPEINFO_FIELD._serialized_end=5856
  _TYPEINFO_TUPLETYPEINFO._serialized_start=5858
  _TYPEINFO_TUPLETYPEINFO._serialized_end=5938
  _TYPEINFO_AVROTYPEINFO._serialized_start=5940
  _TYPEINFO_AVROTYPEINFO._serialized_end=5970
  _TYPEINFO_TYPENAME._serialized_start=5973
  _TYPEINFO_TYPENAME._serialized_end=6370
  _USERDEFINEDDATASTREAMFUNCTION._serialized_start=6386
  _USERDEFINEDDATASTREAMFUNCTION._serialized_end=7363
  _USERDEFINEDDATASTREAMFUNCTION_RUNTIMECONTEXT._serialized_start=6881
  _USERDEFINEDDATASTREAMFUNCTION_RUNTIMECONTEXT._serialized_end=7187
  _USERDEFINEDDATASTREAMFUNCTION_FUNCTIONTYPE._serialized_start=7190
  _USERDEFINEDDATASTREAMFUNCTION_FUNCTIONTYPE._serialized_end=7363
  _STATEDESCRIPTOR._serialized_start=7366
  _STATEDESCRIPTOR._serialized_end=9258
  _STATEDESCRIPTOR_STATETTLCONFIG._serialized_start=7498
  _STATEDESCRIPTOR_STATETTLCONFIG._serialized_end=9258
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES._serialized_start=7969
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES._serialized_end=9067
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_INCREMENTALCLEANUPSTRATEGY._serialized_start=8147
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_INCREMENTALCLEANUPSTRATEGY._serialized_end=8235
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_ROCKSDBCOMPACTFILTERCLEANUPSTRATEGY._serialized_start=8237
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_ROCKSDBCOMPACTFILTERCLEANUPSTRATEGY._serialized_end=8312
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_MAPSTRATEGIESENTRY._serialized_start=8315
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_MAPSTRATEGIESENTRY._serialized_end=8923
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_STRATEGIES._serialized_start=8925
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_STRATEGIES._serialized_end=9023
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_EMPTYCLEANUPSTRATEGY._serialized_start=9025
  _STATEDESCRIPTOR_STATETTLCONFIG_CLEANUPSTRATEGIES_EMPTYCLEANUPSTRATEGY._serialized_end=9067
  _STATEDESCRIPTOR_STATETTLCONFIG_UPDATETYPE._serialized_start=9069
  _STATEDESCRIPTOR_STATETTLCONFIG_UPDATETYPE._serialized_end=9137
  _STATEDESCRIPTOR_STATETTLCONFIG_STATEVISIBILITY._serialized_start=9139
  _STATEDESCRIPTOR_STATETTLCONFIG_STATEVISIBILITY._serialized_end=9213
  _STATEDESCRIPTOR_STATETTLCONFIG_TTLTIMECHARACTERISTIC._serialized_start=9215
  _STATEDESCRIPTOR_STATETTLCONFIG_TTLTIMECHARACTERISTIC._serialized_end=9258
  _CODERINFODESCRIPTOR._serialized_start=9261
  _CODERINFODESCRIPTOR._serialized_end=10270
  _CODERINFODESCRIPTOR_FLATTENROWTYPE._serialized_start=9854
  _CODERINFODESCRIPTOR_FLATTENROWTYPE._serialized_end=9928
  _CODERINFODESCRIPTOR_ROWTYPE._serialized_start=9930
  _CODERINFODESCRIPTOR_ROWTYPE._serialized_end=9997
  _CODERINFODESCRIPTOR_ARROWTYPE._serialized_start=9999
  _CODERINFODESCRIPTOR_ARROWTYPE._serialized_end=10068
  _CODERINFODESCRIPTOR_OVERWINDOWARROWTYPE._serialized_start=10070
  _CODERINFODESCRIPTOR_OVERWINDOWARROWTYPE._serialized_end=10149
  _CODERINFODESCRIPTOR_RAWTYPE._serialized_start=10151
  _CODERINFODESCRIPTOR_RAWTYPE._serialized_end=10223
  _CODERINFODESCRIPTOR_MODE._serialized_start=10225
  _CODERINFODESCRIPTOR_MODE._serialized_end=10257
# @@protoc_insertion_point(module_scope)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import warnings
from abc import abstractmethod

from typing import TYPE_CHECKING, Optional

from pyflink.common.serialization import BulkWriterFactory, RowDataBulkWriterFactory

if TYPE_CHECKING:
    from pyflink.table.types import RowType

from pyflink.common import Duration, Encoder
from pyflink.datastream.connectors import Source, Sink
from pyflink.datastream.connectors.base import SupportsPreprocessing, StreamTransformer
from pyflink.datastream.functions import SinkFunction
from pyflink.common.utils import JavaObjectWrapper
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import to_jarray, is_instance_of

__all__ = [
    'FileCompactor',
    'FileCompactStrategy',
    'OutputFileConfig',
    'FileSource',
    'FileSourceBuilder',
    'FileSink',
    'StreamingFileSink',
    'StreamFormat',
    'BulkFormat',
    'FileEnumeratorProvider',
    'FileSplitAssignerProvider',
    'RollingPolicy',
    'BucketAssigner'
]


# ---- FileSource ----

class FileEnumeratorProvider(object):
    """
    Factory for FileEnumerator which task is to discover all files to be read and to split them
    into a set of file source splits. This includes possibly, path traversals, file filtering
    (by name or other patterns) and deciding whether to split files into multiple splits, and
    how to split them.
    """

    def __init__(self, j_file_enumerator_provider):
        self._j_file_enumerator_provider = j_file_enumerator_provider

    @staticmethod
    def default_splittable_file_enumerator() -> 'FileEnumeratorProvider':
        """
        The default file enumerator used for splittable formats. The enumerator recursively
        enumerates files, split files that consist of multiple distributed storage blocks into
        multiple splits, and filters hidden files (files starting with '.' or '_'). Files with
        suffixes of common compression formats (for example '.gzip', '.bz2', '.xy', '.zip', ...)
        will not be split.
        """
        JFileSource = get_gateway().jvm.org.apache.flink.connector.file.src.FileSource
        return FileEnumeratorProvider(JFileSource.DEFAULT_SPLITTABLE_FILE_ENUMERATOR)

    @staticmethod
    def default_non_splittable_file_enumerator() -> 'FileEnumeratorProvider':
        """
        The default file enumerator used for non-splittable formats. The enumerator recursively
        enumerates files, creates one split for the file, and filters hidden files
        (files starting with '.' or '_').
        """
        JFileSource = get_gateway().jvm.org.apache.flink.connector.file.src.FileSource
        return FileEnumeratorProvider(JFileSource.DEFAULT_NON_SPLITTABLE_FILE_ENUMERATOR)


class FileSplitAssignerProvider(object):
    """
    Factory for FileSplitAssigner which is responsible for deciding what split should be
    processed next by which node. It determines split processing order and locality.
    """

    def __init__(self, j_file_split_assigner):
        self._j_file_split_assigner = j_file_split_assigner

    @staticmethod
    def locality_aware_split_assigner() -> 'FileSplitAssignerProvider':
        """
        A FileSplitAssigner that assigns to each host preferably splits that are local, before
        assigning splits that are not local.
        """
        JFileSource = get_gateway().jvm.org.apache.flink.connector.file.src.FileSource
        return FileSplitAssignerProvider(JFileSource.DEFAULT_SPLIT_ASSIGNER)


class StreamFormat(object):
    """
    A reader format that reads individual records from a stream.

    Compared to the :class:`~BulkFormat`, the stream format handles a few things out-of-the-box,
    like deciding how to batch records or dealing with compression.

    Internally in the file source, the readers pass batches of records from the reading threads
    (that perform the typically blocking I/O operations) to the async mailbox threads that do
    the streaming and batch data processing. Passing records in batches
    (rather than one-at-a-time) much reduces the thread-to-thread handover overhead.

    This batching is by default based on I/O fetch size for the StreamFormat, meaning the
    set of records derived from one I/O buffer will be handed over as one. See config option
    `source.file.stream.io-fetch-size` to configure that fetch size.
    """

    def __init__(self, j_stream_format):
        self._j_stream_format = j_stream_format

    @staticmethod
    def text_line_format(charset_name: str = "UTF-8") -> 'StreamFormat':
        """
        Creates a reader format that text lines from a file.

        The reader uses Java's built-in java.io.InputStreamReader to decode the byte stream
        using various supported charset encodings.

        This format does not support optimized recovery from checkpoints. On recovery, it will
        re-read and discard the number of lined that were processed before the last checkpoint.
        That is due to the fact that the offsets of lines in the file cannot be tracked through
        the charset decoders with their internal buffering of stream input and charset decoder
        state.

        :param charset_name: The charset to decode the byte stream.
        """
        j_stream_format = get_gateway().jvm.org.apache.flink.connector.file.src.reader. \
            TextLineInputFormat(charset_name)
        return StreamFormat(j_stream_format)


class BulkFormat(object):
    """
    The BulkFormat reads and decodes batches of records at a time. Examples of bulk formats are
    formats like ORC or Parquet.

    Internally in the file source, the readers pass batches of records from the reading threads
    (that perform the typically blocking I/O operations) to the async mailbox threads that do the
    streaming and batch data processing. Passing records in batches (rather than one-at-a-time) much
    reduce the thread-to-thread handover overhead.

    For the BulkFormat, one batch is handed over as one.

    .. versionadded:: 1.16.0
    """

    def __init__(self, j_bulk_format):
        self._j_bulk_format = j_bulk_format


class FileSourceBuilder(object):
    """
    The builder for the :class:`~FileSource`, to configure the various behaviors.

    Start building the source via one of the following methods:

        - :func:`~FileSource.for_record_stream_format`
    """

    def __init__(self, j_file_source_builder):
        self._j_file_source_builder = j_file_source_builder

    def monitor_continuously(
            self,
            discovery_interval: Duration) -> 'FileSourceBuilder':
        """
        Sets this source to streaming ("continuous monitoring") mode.

        This makes the source a "continuous streaming" source that keeps running, monitoring
        for new files, and reads these files when they appear and are discovered by the
        monitoring.

        The interval in which the source checks for new files is the discovery_interval. Shorter
        intervals mean that files are discovered more quickly, but also imply more frequent
        listing or directory traversal of the file system / object store.
        """
        self._j_file_source_builder.monitorContinuously(discovery_interval._j_duration)
        return self

    def process_static_file_set(self) -> 'FileSourceBuilder':
        """
        Sets this source to bounded (batch) mode.

        In this mode, the source processes the files that are under the given paths when the
        application is started. Once all files are processed, the source will finish.

        This setting is also the default behavior. This method is mainly here to "switch back"
        to bounded (batch) mode, or to make it explicit in the source construction.
        """
        self._j_file_source_builder.processStaticFileSet()
        return self

    def set_file_enumerator(
            self,
            file_enumerator: 'FileEnumeratorProvider') -> 'FileSourceBuilder':
        """
        Configures the FileEnumerator for the source. The File Enumerator is responsible
        for selecting from the input path the set of files that should be processed (and which
        to filter out). Furthermore, the File Enumerator may split the files further into
        sub-regions, to enable parallelization beyond the number of files.
        """
        self._j_file_source_builder.setFileEnumerator(
            file_enumerator._j_file_enumerator_provider)
        return self

    def set_split_assigner(
            self,
            split_assigner: 'FileSplitAssignerProvider') -> 'FileSourceBuilder':
        """
        Configures the FileSplitAssigner for the source. The File Split Assigner
        determines which parallel reader instance gets which {@link FileSourceSplit}, and in
        which order these splits are assigned.
        """
        self._j_file_source_builder.setSplitAssigner(split_assigner._j_file_split_assigner)
        return self

    def build(self) -> 'FileSource':
        """
        Creates the file source with the settings applied to this builder.
        """
        return FileSource(self._j_file_source_builder.build())


class FileSource(Source):
    """
    A unified data source that reads files - both in batch and in streaming mode.

    This source supports all (distributed) file systems and object stores that can be accessed via
    the Flink's FileSystem class.

    Start building a file source via one of the following calls:

        - :func:`~FileSource.for_record_stream_format`

    This creates a :class:`~FileSource.FileSourceBuilder` on which you can configure all the
    properties of the file source.

    <h2>Batch and Streaming</h2>

    This source supports both bounded/batch and continuous/streaming data inputs. For the
    bounded/batch case, the file source processes all files under the given path(s). In the
    continuous/streaming case, the source periodically checks the paths for new files and will start
    reading those.

    When you start creating a file source (via the
    :class:`~FileSource.FileSourceBuilder` created through one of the above-mentioned methods)
    the source is by default in bounded/batch mode. Call
    :func:`~FileSource.FileSourceBuilder.monitor_continuously` to put the source into continuous
    streaming mode.

    <h2>Format Types</h2>

    The reading of each file happens through file readers defined by <i>file formats</i>. These
    define the parsing logic for the contents of the file. There are multiple classes that the
    source supports. Their interfaces trade of simplicity of implementation and
    flexibility/efficiency.

        - A :class:`~FileSource.StreamFormat` reads the contents of a file from a file stream.
          It is the simplest format to implement, and provides many features out-of-the-box
          (like checkpointing logic) but is limited in the optimizations it
          can apply (such as object reuse, batching, etc.).

    <h2>Discovering / Enumerating Files</h2>

    The way that the source lists the files to be processes is defined by the
    :class:`~FileSource.FileEnumeratorProvider`. The FileEnumeratorProvider is responsible to
    select the relevant files (for example filter out hidden files) and to optionally splits files
    into multiple regions (= file source splits) that can be read in parallel).
    """

    def __init__(self, j_file_source):
        super(FileSource, self).__init__(source=j_file_source)

    @staticmethod
    def for_record_stream_format(stream_format: StreamFormat, *paths: str) -> FileSourceBuilder:
        """
        Builds a new FileSource using a :class:`~FileSource.StreamFormat` to read record-by-record
        from a file stream.

        When possible, stream-based formats are generally easier (preferable) to file-based
        formats, because they support better default behavior around I/O batching or progress
        tracking (checkpoints).

        Stream formats also automatically de-compress files based on the file extension. This
        supports files ending in ".deflate" (Deflate), ".xz" (XZ), ".bz2" (BZip2), ".gz", ".gzip"
        (GZip).
        """
        JPath = get_gateway().jvm.org.apache.flink.core.fs.Path
        JFileSource = get_gateway().jvm.org.apache.flink.connector.file.src.FileSource
        j_paths = to_jarray(JPath, [JPath(p) for p in paths])
        return FileSourceBuilder(
            JFileSource.forRecordStreamFormat(stream_format._j_stream_format, j_paths))

    @staticmethod
    def for_bulk_file_format(bulk_format: BulkFormat, *paths: str) -> FileSourceBuilder:
        JPath = get_gateway().jvm.org.apache.flink.core.fs.Path
        JFileSource = get_gateway().jvm.org.apache.flink.connector.file.src.FileSource
        j_paths = to_jarray(JPath, [JPath(p) for p in paths])
        return FileSourceBuilder(
            JFileSource.forBulkFileFormat(bulk_format._j_bulk_format, j_paths))


# ---- FileSink ----


class BucketAssigner(JavaObjectWrapper):
    """
    A BucketAssigner is used with a file sink to determine the bucket each incoming element should
    be put into.

    The StreamingFileSink can be writing to many buckets at a time, and it is responsible
    for managing a set of active buckets. Whenever a new element arrives it will ask the
    BucketAssigner for the bucket the element should fall in. The BucketAssigner can, for
    example, determine buckets based on system time.
    """

    def __init__(self, j_bucket_assigner):
        super().__init__(j_bucket_assigner)

    @staticmethod
    def base_path_bucket_assigner() -> 'BucketAssigner':
        """
        Creates a BucketAssigner that does not perform any bucketing of files. All files are
        written to the base path.
        """
        return BucketAssigner(get_gateway().jvm.org.apache.flink.streaming.api.functions.sink.
                              filesystem.bucketassigners.BasePathBucketAssigner())

    @staticmethod
    def date_time_bucket_assigner(format_str: str = "yyyy-MM-dd--HH", timezone_id: str = None):
        """
        Creates a BucketAssigner that assigns to buckets based on current system time.

        It will create directories of the following form: /{basePath}/{dateTimePath}/}.
        The basePath is the path that was specified as a base path when creating the new bucket.
        The dateTimePath is determined based on the current system time and the user provided format
        string.

        The Java DateTimeFormatter is used to derive a date string from the current system time and
        the date format string. The default format string is "yyyy-MM-dd--HH" so the rolling files
        will have a granularity of hours.

        :param format_str: The format string used to determine the bucket id.
        :param timezone_id: The timezone id, either an abbreviation such as "PST", a full name
                            such as "America/Los_Angeles", or a custom timezone_id such as
                            "GMT-08:00". Th e default time zone will b used if it's None.
        """
        if timezone_id is not None and isinstance(timezone_id, str):
            j_timezone = get_gateway().jvm.java.time.ZoneId.of(timezone_id)
        else:
            j_timezone = get_gateway().jvm.java.time.ZoneId.systemDefault()
        return BucketAssigner(
            get_gateway().jvm.org.apache.flink.streaming.api.functions.sink.
            filesystem.bucketassigners.DateTimeBucketAssigner(format_str, j_timezone))


class RollingPolicy(JavaObjectWrapper):
    """
    The policy based on which a Bucket in the FileSink rolls its currently
    open part file and opens a new one.
    """

    def __init__(self, j_rolling_policy):
        super().__init__(j_rolling_policy)

    @staticmethod
    def default_rolling_policy(
            part_size: int = 1024 * 1024 * 128,
            rollover_interval: int = 60 * 1000,
            inactivity_interval: int = 60 * 1000) -> 'DefaultRollingPolicy':
        """
        Returns the default implementation of the RollingPolicy.

        This policy rolls a part file if:

            - there is no open part file,
            - the current file has reached the maximum bucket size (by default 128MB),
            - the current file is older than the roll over interval (by default 60 sec), or
            - the current file has not been written to for more than the allowed inactivityTime (by
              default 60 sec).

        :param part_size: The maximum part file size before rolling.
        :param rollover_interval: The maximum time duration a part file can stay open before
                                  rolling.
        :param inactivity_interval: The time duration of allowed inactivity after which a part file
                                    will have to roll.
        """
        JDefaultRollingPolicy = get_gateway().jvm.org.apache.flink.streaming.api.functions.\
            sink.filesystem.rollingpolicies.DefaultRollingPolicy
        j_rolling_policy = JDefaultRollingPolicy.builder()\
            .withMaxPartSize(part_size) \
            .withRolloverInterval(rollover_interval) \
            .withInactivityInterval(inactivity_interval) \
            .build()
        return DefaultRollingPolicy(j_rolling_policy)

    @staticmethod
    def on_checkpoint_rolling_policy() -> 'OnCheckpointRollingPolicy':
        """
        Returns a RollingPolicy which rolls (ONLY) on every checkpoint.
        """
        JOnCheckpointRollingPolicy = get_gateway().jvm.org.apache.flink.streaming.api.functions. \
            sink.filesystem.rollingpolicies.OnCheckpointRollingPolicy
        return OnCheckpointRollingPolicy(JOnCheckpointRollingPolicy.build())


class DefaultRollingPolicy(RollingPolicy):
    """
    The default implementation of the RollingPolicy.

    This policy rolls a part file if:

        - there is no open part file,
        - the current file has reached the maximum bucket size (by default 128MB),
        - the current file is older than the roll over interval (by default 60 sec), or
        - the current file has not been written to for more than the allowed inactivityTime (by
          default 60 sec).
    """

    def __init__(self, j_rolling_policy):
        super().__init__(j_rolling_policy)


class OnCheckpointRollingPolicy(RollingPolicy):
    """
    A RollingPolicy which rolls (ONLY) on every checkpoint.
    """

    def __init__(self, j_rolling_policy):
        super().__init__(j_rolling_policy)


class OutputFileConfig(JavaObjectWrapper):
    """
    Part file name configuration.
    This allow to define a prefix and a suffix to the part file name.
    """

    @staticmethod
    def builder():
        return OutputFileConfig.OutputFileConfigBuilder()

    def __init__(self, part_prefix: str, part_suffix: str):
        filesystem = get_gateway().jvm.org.apache.flink.streaming.api.functions.sink.filesystem
        self._j_output_file_config = filesystem.OutputFileConfig(part_prefix, part_suffix)
        super().__init__(self._j_output_file_config)

    def get_part_prefix(self) -> str:
        """
        The prefix for the part name.
        """
        return self._j_output_file_config.getPartPrefix()

    def get_part_suffix(self) -> str:
        """
        The suffix for the part name.
        """
        return self._j_output_file_config.getPartSuffix()

    class OutputFileConfigBuilder(object):
        """
        A builder to create the part file configuration.
        """

        def __init__(self):
            self.part_prefix = "part"
            self.part_suffix = ""

        def with_part_prefix(self, prefix) -> 'OutputFileConfig.OutputFileConfigBuilder':
            self.part_prefix = prefix
            return self

        def with_part_suffix(self, suffix) -> 'OutputFileConfig.OutputFileConfigBuilder':
            self.part_suffix = suffix
            return self

        def build(self) -> 'OutputFileConfig':
            return OutputFileConfig(self.part_prefix, self.part_suffix)


class FileCompactStrategy(JavaObjectWrapper):
    """
    Strategy for compacting the files written in {@link FileSink} before committing.

    .. versionadded:: 1.16.0
    """

    def __init__(self, j_file_compact_strategy):
        super().__init__(j_file_compact_strategy)

    @staticmethod
    def builder() -> 'FileCompactStrategy.Builder':
        return FileCompactStrategy.Builder()

    class Builder(object):

        def __init__(self):
            JFileCompactStrategy = get_gateway().jvm.org.apache.flink.connector.file.sink.\
                compactor.FileCompactStrategy
            self._j_builder = JFileCompactStrategy.Builder.newBuilder()

        def build(self) -> 'FileCompactStrategy':
            return FileCompactStrategy(self._j_builder.build())

        def enable_compaction_on_checkpoint(self, num_checkpoints_before_compaction: int) \
                -> 'FileCompactStrategy.Builder':
            """
            Optional, compaction will be triggered when N checkpoints passed since the last
            triggering, -1 by default indicating no compaction on checkpoint.
            """
            self._j_builder.enableCompactionOnCheckpoint(num_checkpoints_before_compaction)
            return self

        def set_size_threshold(self, size_threshold: int) -> 'FileCompactStrategy.Builder':
            """
            Optional, compaction will be triggered when the total size of compacting files reaches
            the threshold. -1 by default, indicating the size is unlimited.
            """
            self._j_builder.setSizeThreshold(size_threshold)
            return self

        def set_num_compact_threads(self, num_compact_threads: int) \
                -> 'FileCompactStrategy.Builder':
            """
            Optional, the count of compacting threads in a compactor operator, 1 by default.
            """
            self._j_builder.setNumCompactThreads(num_compact_threads)
            return self


class FileCompactor(JavaObjectWrapper):
    """
    The FileCompactor is responsible for compacting files into one file.

    .. versionadded:: 1.16.0
    """

    def __init__(self, j_file_compactor):
        super().__init__(j_file_compactor)

    @staticmethod
    def concat_file_compactor(file_delimiter: bytes = None):
        """
        Returns a file compactor that simply concat the compacting files. The file_delimiter will be
        added between neighbouring files if provided.
        """
        JConcatFileCompactor = get_gateway().jvm.org.apache.flink.connector.file.sink.compactor.\
            ConcatFileCompactor
        if file_delimiter:
            return FileCompactor(JConcatFileCompactor(file_delimiter))
        else:
            return FileCompactor(JConcatFileCompactor())

    @staticmethod
    def identical_file_compactor():
        """
        Returns a file compactor that directly copy the content of the only input file to the
        output.
        """
        JIdenticalFileCompactor = get_gateway().jvm.org.apache.flink.connector.file.sink.compactor.\
            IdenticalFileCompactor
        return FileCompactor(JIdenticalFileCompactor())


class FileSink(Sink, SupportsPreprocessing):
    """
    A unified sink that emits its input elements to FileSystem files within buckets. This
    sink achieves exactly-once semantics for both BATCH and STREAMING.

    When creating the sink a basePath must be specified. The base directory contains one
    directory for every bucket. The bucket directories themselves contain several part files, with
    at least one for each parallel subtask of the sink which is writing data to that bucket.
    These part files contain the actual output data.

    The sink uses a BucketAssigner to determine in which bucket directory each element
    should be written to inside the base directory. The BucketAssigner can, for example, roll
    on every checkpoint or use time or a property of the element to determine the bucket directory.
    The default BucketAssigner is a DateTimeBucketAssigner which will create one new
    bucket every hour. You can specify a custom BucketAssigner using the
    :func:`~FileSink.RowFormatBuilder.with_bucket_assigner`, after calling
    :class:`~FileSink.for_row_format`.

    The names of the part files could be defined using OutputFileConfig. This
    configuration contains a part prefix and a part suffix that will be used with a random uid
    assigned to each subtask of the sink and a rolling counter to determine the file names. For
    example with a prefix "prefix" and a suffix ".ext", a file named {@code
    "prefix-81fc4980-a6af-41c8-9937-9939408a734b-17.ext"} contains the data from subtask with uid
    {@code 81fc4980-a6af-41c8-9937-9939408a734b} of the sink and is the {@code 17th} part-file
    created by that subtask.

    Part files roll based on the user-specified RollingPolicy. By default, a DefaultRollingPolicy
    is used for row-encoded sink output; a OnCheckpointRollingPolicy is
    used for bulk-encoded sink output.

    In some scenarios, the open buckets are required to change based on time. In these cases, the
    user can specify a bucket_check_interval (by default 1m) and the sink will check
    periodically and roll the part file if the specified rolling policy says so.

    Part files can be in one of three states: in-progress, pending or finished. The reason for this
    is how the sink works to provide exactly-once semantics and fault-tolerance. The part file that
    is currently being written to is in-progress. Once a part file is closed for writing it becomes
    pending. When a checkpoint is successful (for STREAMING) or at the end of the job (for BATCH)
    the currently pending files will be moved to finished.

    For STREAMING in order to guarantee exactly-once semantics in case of a failure, the
    sink should roll back to the state it had when that last successful checkpoint occurred. To this
    end, when restoring, the restored files in pending state are transferred into the finished state
    while any in-progress files are rolled back, so that they do not contain data that arrived after
    the checkpoint from which we restore.
    """

    def __init__(self, j_file_sink, transformer: Optional[StreamTransformer] = None):
        super(FileSink, self).__init__(sink=j_file_sink)
        self._transformer = transformer

    def get_transformer(self) -> Optional[StreamTransformer]:
        return self._transformer

    class BaseBuilder(object):

        def __init__(self, j_builder):
            self._j_builder = j_builder

        def with_bucket_check_interval(self, interval: int):
            """
            :param interval: The check interval in milliseconds.
            """
            self._j_builder.withBucketCheckInterval(interval)
            return self

        def with_bucket_assigner(self, bucket_assigner: BucketAssigner):
            self._j_builder.withBucketAssigner(bucket_assigner.get_java_object())
            return self

        def with_output_file_config(self, output_file_config: OutputFileConfig):
            self._j_builder.withOutputFileConfig(output_file_config.get_java_object())
            return self

        def enable_compact(self, strategy: FileCompactStrategy, compactor: FileCompactor):
            self._j_builder.enableCompact(strategy.get_java_object(), compactor.get_java_object())
            return self

        def disable_compact(self):
            self._j_builder.disableCompact()
            return self

        @abstractmethod
        def with_rolling_policy(self, rolling_policy):
            pass

        def build(self):
            return FileSink(self._j_builder.build())

    class RowFormatBuilder(BaseBuilder):
        """
        Builder for the vanilla FileSink using a row format.

        .. versionchanged:: 1.16.0
           Support compaction.
        """

        def __init__(self, j_row_format_builder):
            super().__init__(j_row_format_builder)

        def with_rolling_policy(self, rolling_policy: RollingPolicy):
            self._j_builder.withRollingPolicy(rolling_policy.get_java_object())
            return self

    @staticmethod
    def for_row_format(base_path: str, encoder: Encoder) -> 'FileSink.RowFormatBuilder':
        JPath = get_gateway().jvm.org.apache.flink.core.fs.Path
        JFileSink = get_gateway().jvm.org.apache.flink.connector.file.sink.FileSink

        return FileSink.RowFormatBuilder(
            JFileSink.forRowFormat(JPath(base_path), encoder._j_encoder))

    class BulkFormatBuilder(BaseBuilder):
        """
        Builder for the vanilla FileSink using a bulk format.

        .. versionadded:: 1.16.0
        """

        def __init__(self, j_bulk_format_builder):
            super().__init__(j_bulk_format_builder)
            self._transformer = None

        def with_rolling_policy(self, rolling_policy: OnCheckpointRollingPolicy):
            if not isinstance(rolling_policy, OnCheckpointRollingPolicy):
                raise ValueError('rolling_policy must be OnCheckpointRollingPolicy for bulk format')
            return self

        def _with_row_type(self, row_type: 'RowType') -> 'FileSink.BulkFormatBuilder':
            from pyflink.datastream.data_stream import DataStream
            from pyflink.table.types import _to_java_data_type

            def _check_if_row_data_type(ds) -> bool:
                j_type_info = ds._j_data_stream.getType()
                if not is_instance_of(
                    j_type_info,
                    'org.apache.flink.table.runtime.typeutils.InternalTypeInfo'
                ):
                    return False
                return is_instance_of(
                    j_type_info.toLogicalType(),
                    'org.apache.flink.table.types.logical.RowType'
                )

            class RowRowTransformer(StreamTransformer):

                def apply(self, ds):
                    jvm = get_gateway().jvm

                    if _check_if_row_data_type(ds):
                        return ds

                    j_map_function = jvm.org.apache.flink.python.util.PythonConnectorUtils \
                        .RowRowMapper(_to_java_data_type(row_type))
                    return DataStream(ds._j_data_stream.process(j_map_function))

            self._transformer = RowRowTransformer()
            return self

        def build(self) -> 'FileSink':
            return FileSink(self._j_builder.build(), self._transformer)

    @staticmethod
    def for_bulk_format(base_path: str, writer_factory: BulkWriterFactory) \
            -> 'FileSink.BulkFormatBuilder':
        jvm = get_gateway().jvm
        j_path = jvm.org.apache.flink.core.fs.Path(base_path)
        JFileSink = jvm.org.apache.flink.connector.file.sink.FileSink

        builder = FileSink.BulkFormatBuilder(
            JFileSink.forBulkFormat(j_path, writer_factory.get_java_object())
        )
        if isinstance(writer_factory, RowDataBulkWriterFactory):
            return builder._with_row_type(writer_factory.get_row_type())
        else:
            return builder


# ---- StreamingFileSink ----


class StreamingFileSink(SinkFunction):
    """
    Sink that emits its input elements to `FileSystem` files within buckets. This is
    integrated with the checkpointing mechanism to provide exactly once semantics.


    When creating the sink a `basePath` must be specified. The base directory contains
    one directory for every bucket. The bucket directories themselves contain several part files,
    with at least one for each parallel subtask of the sink which is writing data to that bucket.
    These part files contain the actual output data.
    """

    def __init__(self, j_obj):
        warnings.warn("Deprecated in 1.15. Use FileSink instead.", DeprecationWarning)
        super(StreamingFileSink, self).__init__(j_obj)

    class BaseBuilder(object):

        def __init__(self, j_builder):
            self._j_builder = j_builder

        def with_bucket_check_interval(self, interval: int):
            self._j_builder.withBucketCheckInterval(interval)
            return self

        def with_bucket_assigner(self, bucket_assigner: BucketAssigner):
            self._j_builder.withBucketAssigner(bucket_assigner.get_java_object())
            return self

        @abstractmethod
        def with_rolling_policy(self, policy):
            pass

        def with_output_file_config(self, output_file_config: OutputFileConfig):
            self._j_builder.withOutputFileConfig(output_file_config.get_java_object())
            return self

        def build(self) -> 'StreamingFileSink':
            j_stream_file_sink = self._j_builder.build()
            return StreamingFileSink(j_stream_file_sink)

    class DefaultRowFormatBuilder(BaseBuilder):
        """
        Builder for the vanilla `StreamingFileSink` using a row format.
        """

        def __init__(self, j_default_row_format_builder):
            super().__init__(j_default_row_format_builder)

        def with_rolling_policy(self, policy: RollingPolicy):
            self._j_builder.withRollingPolicy(policy.get_java_object())
            return self

    @staticmethod
    def for_row_format(base_path: str, encoder: Encoder) -> 'DefaultRowFormatBuilder':
        j_path = get_gateway().jvm.org.apache.flink.core.fs.Path(base_path)
        j_default_row_format_builder = get_gateway().jvm.org.apache.flink.streaming.api.\
            functions.sink.filesystem.legacy.StreamingFileSink.\
            forRowFormat(j_path, encoder._j_encoder)
        return StreamingFileSink.DefaultRowFormatBuilder(j_default_row_format_builder)

    class DefaultBulkFormatBuilder(BaseBuilder):

        def __init__(self, j_default_bulk_format_builder):
            super().__init__(j_default_bulk_format_builder)

        def with_rolling_policy(self, policy: OnCheckpointRollingPolicy):
            self._j_builder.withRollingPolicy(policy.get_java_object())
            return self

    @staticmethod
    def for_bulk_format(base_path: str, writer_factory: BulkWriterFactory):
        jvm = get_gateway().jvm
        j_path = jvm.org.apache.flink.core.fs.Path(base_path)
        j_default_bulk_format_builder = jvm.org.apache.flink.streaming.api.functions.sink \
            .filesystem.legacy.StreamingFileSink.\
            forBulkFormat(j_path, writer_factory.get_java_object())
        return StreamingFileSink.DefaultBulkFormatBuilder(j_default_bulk_format_builder)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import re
import sys

# force to register the operations to SDK Harness
from apache_beam.options.pipeline_options import DebugOptions, PipelineOptions

import pyflink.fn_execution.beam.beam_operations # noqa # pylint: disable=unused-import

# force to register the coders to SDK Harness
import pyflink.fn_execution.beam.beam_coders # noqa # pylint: disable=unused-import

import apache_beam

# disable bundle processor shutdown
from apache_beam.runners.worker import sdk_worker, sdk_worker_main, statecache
sdk_worker.DEFAULT_BUNDLE_PROCESSOR_CACHE_SHUTDOWN_THRESHOLD_S = 86400 * 30


# Currently PyFlink only support count-based state cache strategy
def get_deep_size(*objs):
    return 1


def get_state_cache_size(options):
    """
    Return the maximum size of state cache in count.
    """
    if isinstance(options, PipelineOptions):
        experiments = options.view_as(DebugOptions).experiments or []
    else:
        experiments = options

    for experiment in experiments:
        # There should only be 1 match so returning from the loop
        if re.match(r'state_cache_size=', experiment):
            return int(
                re.match(r'state_cache_size=(?P<state_cache_size>.*)',
                         experiment).group('state_cache_size'))
    return 0


statecache.get_deep_size = get_deep_size
sdk_worker_main._get_state_cache_size = get_state_cache_size
# since Beam 2.52.0, _get_state_cache_size is renamed to _get_state_cache_size_bytes
sdk_worker_main._get_state_cache_size_bytes = get_state_cache_size


def print_to_logging(logging_func, msg, *args, **kwargs):
    if msg != '\n':
        logging_func(msg, *args, **kwargs)


class CustomPrint(object):
    def __init__(self, _print):
        self._msg_buffer = []
        self._print = _print

    def print(self, *args, sep=' ', end='\n', file=None):
        self._msg_buffer.append(sep.join([str(arg) for arg in args]))
        if end == '\n':
            self._print(''.join(self._msg_buffer), sep=sep, end=end, file=file)
            self._msg_buffer.clear()
        else:
            self._msg_buffer.append(end)

    def close(self):
        if self._msg_buffer:
            self._print(''.join(self._msg_buffer), sep='', end='\n')
            self._msg_buffer.clear()


def main():
    import builtins
    import logging
    from functools import partial

    # redirect stdout to logging.info, stderr to logging.error
    _info = logging.getLogger().info
    _error = logging.getLogger().error
    sys.stdout.write = partial(print_to_logging, _info)
    sys.stderr.write = partial(print_to_logging, _error)

    custom_print = CustomPrint(print)
    builtins.print = custom_print.print
    # Remove all the built-in log handles
    logging.getLogger().handlers = []
    apache_beam.runners.worker.sdk_worker_main.main(sys.argv)
    custom_print.close()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# cython: language_level = 3
# cython: infer_types = True
# cython: profile=True
# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True

from apache_beam.coders.coder_impl cimport OutputStream as BOutputStream
from apache_beam.coders.coder_impl cimport StreamCoderImpl

from pyflink.fn_execution.coder_impl_fast cimport LengthPrefixBaseCoderImpl, FieldCoderImpl
from pyflink.fn_execution.beam.beam_stream_fast cimport BeamTimeBasedOutputStream
from pyflink.fn_execution.stream_fast cimport OutputStream

cdef class PassThroughLengthPrefixCoderImpl(StreamCoderImpl):
    cdef readonly StreamCoderImpl _value_coder

cdef class FlinkFieldCoderBeamWrapper(StreamCoderImpl):
    cdef FieldCoderImpl _value_coder
    cdef OutputStream _data_out_stream

    cdef void _write_data_output_stream(self, BOutputStream out_stream)

cdef class FlinkLengthPrefixCoderBeamWrapper(StreamCoderImpl):
    cdef readonly LengthPrefixBaseCoderImpl _value_coder
    cdef readonly BeamTimeBasedOutputStream _output_stream



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from abc import ABCMeta

from py4j.java_gateway import get_java_class
from typing import Optional

from pyflink.java_gateway import get_gateway

__all__ = [
    'CheckpointStorage',
    'JobManagerCheckpointStorage',
    'FileSystemCheckpointStorage',
    'CustomCheckpointStorage']


def _from_j_checkpoint_storage(j_checkpoint_storage):
    if j_checkpoint_storage is None:
        return None
    gateway = get_gateway()
    JCheckpointStorage = gateway.jvm.org.apache.flink.runtime.state.CheckpointStorage
    JJobManagerCheckpointStorage = gateway.jvm.org.apache.flink.runtime.state.storage \
        .JobManagerCheckpointStorage
    JFileSystemCheckpointStorage = gateway.jvm.org.apache.flink.runtime.state.storage \
        .FileSystemCheckpointStorage

    j_clz = j_checkpoint_storage.getClass()

    if not get_java_class(JCheckpointStorage).isAssignableFrom(j_clz):
        raise TypeError("%s is not an instance of CheckpointStorage." % j_checkpoint_storage)

    if get_java_class(JJobManagerCheckpointStorage).isAssignableFrom(j_clz):
        return JobManagerCheckpointStorage(j_jobmanager_checkpoint_storage=j_checkpoint_storage)
    elif get_java_class(JFileSystemCheckpointStorage).isAssignableFrom(j_clz):
        return FileSystemCheckpointStorage(j_filesystem_checkpoint_storage=j_checkpoint_storage)
    else:
        return CustomCheckpointStorage(j_checkpoint_storage)


class CheckpointStorage(object, metaclass=ABCMeta):
    """
    Checkpoint storage defines how :class:`StateBackend`'s store their state for fault-tolerance
    in streaming applications. Various implementations store their checkpoints in different fashions
    and have different requirements and availability guarantees.

    For example, :class:`JobManagerCheckpointStorage` stores checkpoints in the memory of the
    `JobManager`. It is lightweight and without additional dependencies but is not scalable
    and only supports small state sizes. This checkpoints storage policy is convenient for local
    testing and development.

    :class:`FileSystemCheckpointStorage` stores checkpoints in a filesystem. For systems like HDFS
    NFS drives, S3, and GCS, this storage policy supports large state size, in the magnitude of many
    terabytes while providing a highly available foundation for streaming applications. This
    checkpoint storage policy is recommended for most production deployments.

    **Raw Bytes Storage**

    The `CheckpointStorage` creates services for raw bytes storage.

    The raw bytes storage (through the CheckpointStreamFactory) is the fundamental service that
    simply stores bytes in a fault tolerant fashion. This service is used by the JobManager to
    store checkpoint and recovery metadata and is typically also used by the keyed- and operator-
    state backends to store checkpoint state.

    **Serializability**

    Implementations need to be serializable(`java.io.Serializable`), because they are distributed
    across parallel processes (for distributed execution) together with the streaming application
    code.

    Because of that `CheckpointStorage` implementations are meant to be like _factories_ that create
    the proper state stores that provide access to the persistent layer. That way, the storage
    policy can be very lightweight (contain only configurations) which makes it easier to be
    serializable.

    **Thread Safety**

    Checkpoint storage implementations have to be thread-safe. Multiple threads may be creating
    streams concurrently.
    """

    def __init__(self, j_checkpoint_storage):
        self._j_checkpoint_storage = j_checkpoint_storage


class JobManagerCheckpointStorage(CheckpointStorage):
    """
    The `CheckpointStorage` checkpoints state directly to the JobManager's memory (hence the
    name), but savepoints will be persisted to a file system.

    This checkpoint storage is primarily for experimentation, quick local setups, or for streaming
    applications that have very small state: Because it requires checkpoints to go through the
    JobManager's memory, larger state will occupy larger portions of the JobManager's main memory,
    reducing operational stability. For any other setup, the `FileSystemCheckpointStorage`
    should be used. The `FileSystemCheckpointStorage` but checkpoints state directly to files
    rather than to the JobManager's memory, thus supporting larger state sizes and more highly
    available recovery.

    **State Size Considerations**

    State checkpointing with this checkpoint storage is subject to the following conditions:

    - Each individual state must not exceed the configured maximum state size
      (see :func:`get_max_state_size`.

    - All state from one task (i.e., the sum of all operator states and keyed states from all
      chained operators of the task) must not exceed what the RPC system supports, which is
      be default < 10 MB. That limit can be configured up, but that is typically not advised.

    - The sum of all states in the application times all retained checkpoints must comfortably
      fit into the JobManager's JVM heap space.

    **Persistence Guarantees**

    For the use cases where the state sizes can be handled by this storage, it does
    guarantee persistence for savepoints, externalized checkpoints (of configured), and checkpoints
    (when high-availability is configured).

        **Configuration**

    As for all checkpoint storage, this type can either be configured within the application (by
    creating the storage with the respective constructor parameters and setting it on the execution
    environment) or by specifying it in the Flink configuration.

    If the storage was specified in the application, it may pick up additional configuration
    parameters from the Flink configuration. For example, if the backend if configured in the
    application without a default savepoint directory, it will pick up a default savepoint
    directory specified in the Flink configuration of the running job/cluster. That behavior is
    implemented via the :func:`configure` method.
    """

    # The default maximal size that the snapshotted memory state may have (5 MiBytes).
    DEFAULT_MAX_STATE_SIZE = 5 * 1024 * 1024

    def __init__(self,
                 checkpoint_path=None,
                 max_state_size=None,
                 j_jobmanager_checkpoint_storage=None):
        """
        Creates a new JobManagerCheckpointStorage, setting optionally the paths to persist
        checkpoint metadata to, as well as configuring state thresholds.

        WARNING: Increasing the size of this value beyond the default value
        (:data:`DEFAULT_MAX_STATE_SIZE`) should be done with care.
        The checkpointed state needs to be send to the JobManager via limited size RPC messages,
        and there and the JobManager needs to be able to hold all aggregated state in its memory.

        Example:
        ::
            >>> checkpoint_storage = JobManagerCheckpointStorage()

        :param checkpoint_path: The path to write checkpoint metadata to. If none, the value from
                                the runtime configuration will be used.
        :param max_state_size: The maximal size of the serialized state. If none, the
                               :data:`DEFAULT_MAX_STATE_SIZE` will be used.
        :param j_jobmanager_checkpoint_storage: For internal use, please keep none.
        """
        if j_jobmanager_checkpoint_storage is None:
            gateway = get_gateway()
            JJobManagerCheckpointStorage = gateway.jvm.org.apache.flink.runtime.state.storage\
                .JobManagerCheckpointStorage
            JPath = gateway.jvm.org.apache.flink.core.fs.Path

            if checkpoint_path is not None:
                checkpoint_path = JPath(checkpoint_path)
            if max_state_size is None:
                max_state_size = JJobManagerCheckpointStorage.DEFAULT_MAX_STATE_SIZE
            j_jobmanager_checkpoint_storage = JJobManagerCheckpointStorage(checkpoint_path,
                                                                           max_state_size)

        super(JobManagerCheckpointStorage, self).__init__(j_jobmanager_checkpoint_storage)

    def get_checkpoint_path(self) -> Optional[str]:
        """
        Gets the base directory where all the checkpoints are stored.
        The job-specific checkpoint directory is created inside this directory.

        :return: The base directory for checkpoints.
        """
        j_path = self._j_checkpoint_storage.getCheckpointPath()
        if j_path is None:
            return None
        else:
            return j_path.toString()

    def get_max_state_size(self) -> int:
        """
        Gets the maximum size that an individual state can have, as configured in the
        constructor. By default :data:`DEFAULT_MAX_STATE_SIZE` will be used.
        """
        return self._j_checkpoint_storage.getMaxStateSize()

    def get_savepoint_path(self) -> Optional[str]:
        """
        Gets the base directory where all the savepoints are stored.
        The job-specific savepoint directory is created inside this directory.

        :return: The base directory for savepoints.
        """

        j_path = self._j_checkpoint_storage.getSavepointPath()
        if j_path is None:
            return None
        else:
            return j_path.toString()

    def __str__(self):
        return self._j_checkpoint_storage.toString()


class FileSystemCheckpointStorage(CheckpointStorage):
    """
    `FileSystemCheckpointStorage` checkpoints state as files to a filesystem.

    Each checkpoint will store all its files in a subdirectory that includes the
    checkpoints number, such as `hdfs://namenode:port/flink-checkpoints/chk-17/`.

    **State Size Considerations**

    This checkpoint storage stores small state chunks directly with the metadata, to avoid creating
    many small files. The threshold for that is configurable. When increasing this threshold, the
    size of the checkpoint metadata increases. The checkpoint metadata of all retained completed
    checkpoints needs to fit into the JobManager's heap memory. This is typically not a problem,
    unless the threashold `get_min_file_size_threshold` is increased significantly.

    **Persistence Guarantees**

    Checkpoints from this checkpoint storage are as persistent and available as the filesystem
    that it is written to. If the file system is a persistent distributed file system, this
    checkpoint storage supports highly available setups. The backend additionally supports
    savepoints and externalized checkpoints.

    **Configuration**

    As for all checkpoint storage policies, this backend can either be configured within the
    application (by creating the storage with the respective constructor parameters and setting
    it on the execution environment) or by specifying it in the Flink configuration.

    If the checkpoint storage was specified in the application, it may pick up additional
    configuration parameters from the Flink configuration. For example, if the storage is configured
    in the application without a default savepoint directory, it will pick up a default savepoint
    directory specified in the Flink configuration of the running job/cluster.
    """

    # Maximum size of state that is stored with the metadata, rather than in files (1 MiByte).
    MAX_FILE_STATE_THRESHOLD = 1024 * 1024

    def __init__(self,
                 checkpoint_path=None,
                 file_state_size_threshold=None,
                 write_buffer_size=-1,
                 j_filesystem_checkpoint_storage=None):
        """
        Creates a new FileSystemCheckpointStorage, setting the paths for the checkpoint data
        in a file system.

        All file systems for the file system scheme in the URI (e.g., `file://`, `hdfs://`, or
        `s3://`) must be accessible via `FileSystem#get`.

        For a Job targeting HDFS, this means that the URI must either specify the authority (host
        and port), of the Hadoop configuration that describes that information must be in the
        classpath.

        Example:
        ::
            >>> checkpoint_storage = FileSystemCheckpointStorage("hdfs://checkpoints")

        :param checkpoint_path: The path to write checkpoint metadata to. If none, the value from
                                the runtime configuration will be used.
        :param file_state_size_threshold: State below this size will be stored as part of the
                                        metadata, rather than in files. If -1, the value configured
                                        in the runtime configuration will be used, or the default
                                        value (1KB) if nothing is configured.
        :param write_buffer_size: Write buffer size used to serialize state. If -1, the value
                                    configured in the runtime configuration will be used, or the
                                    default value (4KB) if nothing is configured.
        :param j_filesystem_checkpoint_storage: For internal use, please keep none.
        """
        if j_filesystem_checkpoint_storage is None:
            gateway = get_gateway()
            JFileSystemCheckpointStorage = gateway.jvm.org.apache.flink.runtime.state.storage\
                .FileSystemCheckpointStorage
            JPath = gateway.jvm.org.apache.flink.core.fs.Path

            if checkpoint_path is None:
                raise ValueError("checkpoint_path must not be None")
            else:
                checkpoint_path = JPath(checkpoint_path)

            if file_state_size_threshold is None:
                file_state_size_threshold = -1

            j_filesystem_checkpoint_storage = JFileSystemCheckpointStorage(
                checkpoint_path,
                file_state_size_threshold,
                write_buffer_size)

        super(FileSystemCheckpointStorage, self).__init__(j_filesystem_checkpoint_storage)

    def get_checkpoint_path(self) -> str:
        """
        Gets the base directory where all the checkpoints are stored.
        The job-specific checkpoint directory is created inside this directory.

        :return: The base directory for checkpoints.
        """
        return self._j_checkpoint_storage.getCheckpointPath().toString()

    def get_savepoint_path(self) -> Optional[str]:
        """
        Gets the base directory where all the savepoints are stored.
        The job-specific savepoint directory is created inside this directory.

        :return: The base directory for savepoints.
        """

        j_path = self._j_checkpoint_storage.getSavepointPath()
        if j_path is None:
            return None
        else:
            return j_path.toString()

    def get_min_file_size_threshold(self) -> int:
        """
        Gets the threshold below which state is stored as part of the metadata, rather than in
        file. This threshold ensures the backend does not create a large amount of small files,
        where potentially the file pointers are larget than the state itself.
        """
        return self._j_checkpoint_storage.getMinFileSizeThreshold()

    def get_write_buffer_size(self) -> int:
        """
        Gets the write buffer size for created checkpoint streams.
        """
        return self._j_checkpoint_storage.getWriteBufferSize()

    def __str__(self):
        return self._j_checkpoint_storage.toString()


class CustomCheckpointStorage(CheckpointStorage):
    """
    A wrapper of customized java checkpoint storage.
    """

    def __init__(self, j_custom_checkpoint_storage):
        super(CustomCheckpointStorage, self).__init__(j_custom_checkpoint_storage)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# cython: language_level = 3
# cython: infer_types = True
# cython: profile=True
# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
from libc.stdlib cimport realloc

from apache_beam.coders.coder_impl cimport InputStream as BInputStream
from apache_beam.coders.coder_impl cimport OutputStream as BOutputStream
from apache_beam.coders.coder_impl cimport StreamCoderImpl

from pyflink.fn_execution.beam.beam_stream_fast cimport BeamInputStream
from pyflink.fn_execution.beam.beam_stream_fast cimport BeamTimeBasedOutputStream
from pyflink.fn_execution.stream_fast cimport InputStream

cdef class PassThroughLengthPrefixCoderImpl(StreamCoderImpl):
    def __cinit__(self, value_coder):
        self._value_coder = value_coder

    cpdef encode_to_stream(self, value, BOutputStream out_stream, bint nested):
        self._value_coder.encode_to_stream(value, out_stream, nested)

    cpdef decode_from_stream(self, BInputStream in_stream, bint nested):
        return self._value_coder.decode_from_stream(in_stream, nested)

    cpdef get_estimated_size_and_observables(self, value, bint nested=False):
        return 0, []

cdef class FlinkFieldCoderBeamWrapper(StreamCoderImpl):
    """
    Bridge between Beam coder and Flink coder for the low-level FieldCoder.
    """
    def __cinit__(self, value_coder):
        self._value_coder = value_coder
        self._data_out_stream = OutputStream()

    cpdef encode_to_stream(self, value, BOutputStream out_stream, bint nested):
        self._value_coder.encode_to_stream(value, self._data_out_stream)
        self._write_data_output_stream(out_stream)

    cpdef decode_from_stream(self, BInputStream in_stream, bint nested):
        cdef size_t size
        cdef InputStream data_input_stream

        size = in_stream.size()

        # create InputStream
        data_input_stream = InputStream()
        data_input_stream._input_data = <char*?>in_stream.allc
        data_input_stream._input_pos = in_stream.pos

        result = self._value_coder.decode_from_stream(data_input_stream, size)
        in_stream.pos = data_input_stream._input_pos
        return result

    cdef void _write_data_output_stream(self, BOutputStream out_stream):
        cdef OutputStream data_out_stream
        cdef size_t size, i, pos
        data_out_stream = self._data_out_stream
        size = data_out_stream.pos
        pos = out_stream.pos
        # extend buffer size
        if pos + size > out_stream.buffer_size:
            out_stream.buffer_size += size
            out_stream.data = <char*> realloc(out_stream.data, out_stream.buffer_size)
        # copy data of data output stream to beam output stream
        for i in range(size):
            out_stream.data[pos + i] = data_out_stream.buffer[i]
        out_stream.pos = pos + size

        # clear data output stream
        self._data_out_stream.pos = 0


cdef class FlinkLengthPrefixCoderBeamWrapper(StreamCoderImpl):
    """
    Bridge between Beam coder and Flink coder for the top-level LengthPrefixCoder.
    """
    def __cinit__(self, value_coder):
        self._value_coder = value_coder
        self._output_stream = BeamTimeBasedOutputStream()

    cpdef encode_to_stream(self, value, BOutputStream out_stream, bint nested):
        self._output_stream.reset_output_stream(out_stream)
        self._value_coder.encode_to_stream(value, self._output_stream)

    cpdef decode_from_stream(self, BInputStream in_stream, bint nested):
        cdef BeamInputStream input_stream = BeamInputStream(in_stream, in_stream.size())
        return self._value_coder.decode_from_stream(input_stream)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from apache_beam.coders.coder_impl import create_InputStream, create_OutputStream

from pyflink.fn_execution.stream_slow import InputStream
from pyflink.fn_execution.utils.operation_utils import PeriodicThread


class BeamInputStream(InputStream):
    def __init__(self, input_stream: create_InputStream):
        super(BeamInputStream, self).__init__([])
        self._input_stream = input_stream

    def read(self, size):
        return self._input_stream.read(size)

    def read_byte(self):
        return self._input_stream.read_byte()

    def size(self):
        return self._input_stream.size()


class BeamTimeBasedOutputStream(create_OutputStream):
    def __init__(self):
        super(BeamTimeBasedOutputStream).__init__()
        self._flush_event = False
        self._periodic_flusher = PeriodicThread(1, self.notify_flush)
        self._periodic_flusher.daemon = True
        self._periodic_flusher.start()
        self._output_stream = None

    def write(self, b: bytes):
        self._output_stream.write(b)

    def reset_output_stream(self, output_stream: create_OutputStream):
        self._output_stream = output_stream

    def notify_flush(self):
        self._flush_event = True

    def close(self):
        if self._periodic_flusher:
            self._periodic_flusher.cancel()
            self._periodic_flusher = None

    def maybe_flush(self):
        if self._flush_event:
            self._output_stream.flush()
            self._flush_event = False
        else:
            self._output_stream.maybe_flush()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Any

from apache_beam.coders.coder_impl import StreamCoderImpl, create_InputStream, create_OutputStream

from pyflink.fn_execution.stream_slow import OutputStream
from pyflink.fn_execution.beam.beam_stream_slow import BeamInputStream, BeamTimeBasedOutputStream


class PassThroughLengthPrefixCoderImpl(StreamCoderImpl):
    def __init__(self, value_coder):
        self._value_coder = value_coder

    def encode_to_stream(self, value, out: create_OutputStream, nested: bool) -> Any:
        self._value_coder.encode_to_stream(value, out, nested)

    def decode_from_stream(self, in_stream: create_InputStream, nested: bool) -> Any:
        return self._value_coder.decode_from_stream(in_stream, nested)

    def get_estimated_size_and_observables(self, value: Any, nested=False):
        return 0, []

    def __repr__(self):
        return 'PassThroughLengthPrefixCoderImpl[%s]' % self._value_coder


class FlinkFieldCoderBeamWrapper(StreamCoderImpl):
    """
    Bridge between Beam coder and Flink coder for the low-level FieldCoder.
    """
    def __init__(self, value_coder):
        self._value_coder = value_coder
        self._data_output_stream = OutputStream()

    def encode_to_stream(self, value, out_stream: create_OutputStream, nested):
        self._value_coder.encode_to_stream(value, self._data_output_stream)
        out_stream.write(self._data_output_stream.get())
        self._data_output_stream.clear()

    def decode_from_stream(self, in_stream: create_InputStream, nested):
        data_input_stream = BeamInputStream(in_stream)
        return self._value_coder.decode_from_stream(data_input_stream)

    def __repr__(self):
        return 'FlinkFieldCoderBeamWrapper[%s]' % self._value_coder


class FlinkLengthPrefixCoderBeamWrapper(FlinkFieldCoderBeamWrapper):
    """
    Bridge between Beam coder and Flink coder for the top-level LengthPrefixCoder.
    """
    def __init__(self, value_coder):
        super(FlinkLengthPrefixCoderBeamWrapper, self).__init__(value_coder)
        self._output_stream = BeamTimeBasedOutputStream()

    def encode_to_stream(self, value, out_stream: create_OutputStream, nested):
        self._output_stream.reset_output_stream(out_stream)

        self._value_coder.encode_to_stream(value, self._data_output_stream)
        self._output_stream.write(self._data_output_stream.get())
        self._data_output_stream.clear()

    def __repr__(self):
        return 'FlinkLengthPrefixCoderBeamWrapper[%s]' % self._value_coder



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

__all__ = ['ExplainDetail']


class ExplainDetail(object):
    """
    ExplainDetail defines the types of details for explain result.

    .. versionadded:: 1.11.0
    """

    # The cost information on physical rel node estimated by optimizer.
    # e.g. TableSourceScan(..., cumulative cost = {1.0E8 rows, 1.0E8 cpu, 2.4E9 io, 0.0 network,
    # 0.0 memory}
    ESTIMATED_COST = 0

    # The changelog mode produced by a physical rel node.
    # e.g. GroupAggregate(..., changelogMode=[I,UA,D])
    CHANGELOG_MODE = 1

    # The execution plan in json format of the program.
    JSON_EXECUTION_PLAN = 2

    # The potential risk warnings and SQL optimizer tuning advice analyzed from the physical plan.
    PLAN_ADVICE = 3



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import time

from pyflink.java_gateway import get_gateway


if __name__ == '__main__':
    # just a daemon process used to serve the rpc call from Java.
    gateway = get_gateway()
    watchdog = gateway.jvm.org.apache.flink.client.python.PythonGatewayServer.watchdog
    try:
        while watchdog.ping():
            time.sleep(1)
    finally:
        get_gateway().close()
        exit(0)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import struct


class InputStream(object):
    """
    A pure Python implementation of InputStream.
    """

    def __init__(self, data):
        self.data = data
        self.pos = 0

    def read(self, size):
        self.pos += size
        return self.data[self.pos - size: self.pos]

    def read_byte(self):
        self.pos += 1
        return self.data[self.pos - 1]

    def read_int8(self):
        return struct.unpack('b', self.read(1))[0]

    def read_int16(self):
        return struct.unpack('>h', self.read(2))[0]

    def read_int32(self):
        return struct.unpack('>i', self.read(4))[0]

    def read_int64(self):
        return struct.unpack('>q', self.read(8))[0]

    def read_float(self):
        return struct.unpack('>f', self.read(4))[0]

    def read_double(self):
        return struct.unpack('>d', self.read(8))[0]

    def read_bytes(self):
        size = self.read_int32()
        return self.read(size)

    def read_var_int64(self):
        shift = 0
        result = 0
        while True:
            byte = self.read_byte()
            if byte < 0:
                raise RuntimeError('VarLong not terminated.')

            bits = byte & 0x7F
            if shift >= 64 or (shift >= 63 and bits > 1):
                raise RuntimeError('VarLong too long.')
            result |= bits << shift
            shift += 7
            if not byte & 0x80:
                break
        if result >= 1 << 63:
            result -= 1 << 64
        return result

    def size(self):
        return len(self.data) - self.pos


class OutputStream(object):
    """
    A pure Python implementation of OutputStream.
    """

    def __init__(self):
        self.data = []
        self.byte_count = 0

    def write(self, b: bytes):
        self.data.append(b)
        self.byte_count += len(b)

    def write_byte(self, v):
        self.data.append(chr(v).encode('latin-1'))
        self.byte_count += 1

    def write_int8(self, v: int):
        self.write(struct.pack('b', v))

    def write_int16(self, v: int):
        self.write(struct.pack('>h', v))

    def write_int32(self, v: int):
        self.write(struct.pack('>i', v))

    def write_int64(self, v: int):
        self.write(struct.pack('>q', v))

    def write_float(self, v: float):
        self.write(struct.pack('>f', v))

    def write_double(self, v: float):
        self.write(struct.pack('>d', v))

    def write_bytes(self, v: bytes, size: int):
        self.write_int32(size)
        self.write(v[:size])

    def write_var_int64(self, v: int):
        if v < 0:
            v += 1 << 64
            if v <= 0:
                raise ValueError('Value too large (negative).')
        while True:
            bits = v & 0x7F
            v >>= 7
            if v:
                bits |= 0x80
            self.data.append(chr(bits).encode('latin-1'))
            self.byte_count += 1
            if not v:
                break

    def get(self) -> bytes:
        return b''.join(self.data)

    def size(self) -> int:
        return self.byte_count

    def clear(self):
        self.data.clear()
        self.byte_count = 0
